# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for value.py."""

import collections
import itertools

from absl import logging
from absl.testing import absltest
from absl.testing import parameterized
import tensorflow as tf
from tf_coder import filter_group
from tf_coder import tf_functions
from tf_coder.value_search import function_operation
from tf_coder.value_search import value
from tf_coder.value_search import value_search_settings as settings_module


class DummyValue(value.Value):
  """A dummy implementation of the value.Value ABC for testing purposes."""

  def reconstruct_expression(self):
    return ''


class ValueTest(parameterized.TestCase):

  def test_init_with_primitive(self):
    constant_value = value.ConstantValue(12)
    self.assertEqual(constant_value.type, int)
    self.assertTrue(constant_value.is_primitive)

    self.assertIsNone(constant_value.elem_type)
    self.assertFalse(constant_value.elem_type_is_tensor)
    self.assertIsNone(constant_value.dtype)
    self.assertIsNone(constant_value.shape)

  def test_init_with_dtype(self):
    constant_value = value.ConstantValue(tf.int64)
    self.assertTrue(constant_value.is_dtype)

  @parameterized.named_parameters(
      {
          'testcase_name': 'sequence_of_primitives',
          'sequence': [1.23, 4.56, 7.89],
          'elem_type': float,
          'sequence_dtype': tf.float32,
          'sequence_shape': [3],
      }, {
          'testcase_name': 'sequence_of_tensors',
          'sequence': [tf.constant(1), tf.constant([2.3, 4.5])],
          'elem_type': type(tf.constant(1)),
          'elem_type_is_tensor': True,
      }, {
          'testcase_name': 'sequence_of_sparse_tensors',
          'sequence': [tf.sparse.from_dense(tf.eye(5))],
          'elem_type': tf.SparseTensor,
          'elem_type_is_sparse_tensor': True,
      }, {
          'testcase_name': 'nested_sequence',
          'sequence': [[1, 2], [3, 4], [5, 6]],
          'elem_type': list,
          'sequence_dtype': tf.int32,
          'sequence_shape': [3, 2],
      },
  )
  def test_init_with_sequence(self,
                              sequence,
                              elem_type=None,
                              elem_type_is_tensor=False,
                              elem_type_is_sparse_tensor=False,
                              sequence_dtype=None,
                              sequence_shape=None):
    dummy_value = DummyValue(sequence)
    self.assertTrue(dummy_value.is_sequence)
    self.assertEqual(dummy_value.elem_type, elem_type)
    self.assertEqual(dummy_value.elem_type_is_tensor, elem_type_is_tensor)
    self.assertEqual(dummy_value.elem_type_is_sparse_tensor,
                     elem_type_is_sparse_tensor)
    self.assertEqual(dummy_value.sequence_dtype, sequence_dtype)
    self.assertEqual(dummy_value.sequence_shape, sequence_shape)

  def test_init_with_tensor(self):
    input_value = value.InputValue(tf.constant([[11, 22, 33], [4, 5, 6]]), 'in')
    self.assertTrue(input_value.is_tensor)
    self.assertFalse(input_value.is_sparse_tensor)
    self.assertEqual(input_value.dtype, tf.int32)
    self.assertEqual(input_value.shape, [2, 3])
    # The shape should be a native Python list containing native Python ints.
    self.assertEqual(type(input_value.shape), list)
    self.assertEqual(type(input_value.shape[0]), int)

  def test_init_with_sparse_tensor(self):
    input_value = value.InputValue(
        tf.SparseTensor(
            indices=[[12, 34]],
            values=tf.constant([5.6], dtype=tf.float16),
            dense_shape=[50, 60]),
        'in')
    self.assertTrue(input_value.is_sparse_tensor)
    self.assertFalse(input_value.is_tensor)
    self.assertEqual(input_value.dtype, tf.float16)
    self.assertEqual(input_value.shape, [50, 60])
    # The shape should be a native Python list containing native Python ints.
    self.assertEqual(type(input_value.shape), list)
    self.assertEqual(type(input_value.shape[0]), int)

  @parameterized.named_parameters(
      ('empty_sequence', [], 'Sequences must be nonempty'),
      ('sequence_of_different_types', [1, False],
       'Sequences must contain elements of the same type'),
      ('sequence_of_bad_type', [lambda x: x + 1],
       'Sequence must contain Tensors, SparseTensors, primitives, or other '
       'sequences'),
      ('nested_sequence_not_rectangular', [[1], [2, 3]],
       'Sequence is not tensor-like'),
      ('nested_sequence_with_tensors', [[tf.constant(1)], [tf.constant(2)]],
       'Sequence is not tensor-like'),
      ('nested_sequence_empty_tensor', [[], [], []],
       'Sequence represents an empty tensor'),
      ('tensor_empty', tf.zeros((1, 2, 3))[:0], 'Tensor is empty'),
      ('tensor_too_large', tf.zeros((1, 2, 3, 4, 5, 6)),
       'Tensor value is too large'),
      ('sparse_index_out_of_bounds',
       tf.SparseTensor(indices=[[12, 34]], values=[3], dense_shape=[5, 6]),
       'SparseTensor has out-of-bounds index'),
      ('sparse_too_large',
       tf.SparseTensor(indices=[[0] * 10],
                       values=[3],
                       dense_shape=[5] * 10),
       'SparseTensor value is too large'),
      ('sparse_empty',
       tf.sparse.from_dense(tf.constant([])),
       'SparseTensor is empty'))
  def test_init_raises(self, wrapped_value, regexp):
    with self.assertRaisesRegex(ValueError, regexp):
      DummyValue(wrapped_value)

  def test_repr(self):
    self.assertEqual(repr(DummyValue(tf.constant([1, 2]))), 'tf.int32:[1, 2]')

  def test_hash(self):
    value_1a = value.InputValue(tf.constant([1, 2]), 'a')
    value_1b = value.OutputValue(tf.constant([1, 2]))
    value_2 = value.InputValue(tf.constant([1, 3]), 'a')
    self.assertEqual(hash(value_1a), hash(value_1b))
    self.assertNotEqual(hash(value_1a), hash(value_2))

  def test_eq(self):
    value_1a = value.InputValue(tf.constant([1, 2]), 'a')
    value_1b = value.OutputValue(tf.constant([1, 2]))
    value_2 = value.InputValue(tf.constant([1, 3]), 'a')
    # We want to explicitly test __eq__.
    self.assertTrue(value_1a == value_1b)  # pylint: disable=g-generic-assert
    self.assertFalse(value_1a == value_2)  # pylint: disable=g-generic-assert

  def test_eq_comparing_to_non_value(self):
    input_value = value.InputValue(tf.constant([1, 2]), 'a')
    not_comparable = [1, 2]
    # We want to explicitly test __eq__.
    self.assertFalse(input_value == not_comparable)  # pylint: disable=g-generic-assert

  def test_ne(self):
    value_1a = value.InputValue(tf.constant([1, 2]), 'a')
    value_1b = value.OutputValue(tf.constant([1, 2]))
    value_2 = value.InputValue(tf.constant([1, 3]), 'a')
    # We want to explicitly test __ne__.
    self.assertFalse(value_1a != value_1b)  # pylint: disable=g-generic-assert
    self.assertTrue(value_1a != value_2)  # pylint: disable=g-generic-assert


class OperationValueTest(absltest.TestCase):

  def setUp(self):
    super(OperationValueTest, self).setUp()
    function_info = tf_functions.FunctionInfo(
        name='tf.reduce_max(input_tensor, axis)',
        filter_group=filter_group.FilterGroup.NONE,
        weight=1)
    self.operation = function_operation.FunctionOperation(function_info)
    self.settings = settings_module.default_settings()

  def test_copy(self):
    tensor_value = value.InputValue([[1, 3, 2], [-3, 0, 4]], 'my_input')
    axis_value = value.ConstantValue(1)
    operation_value = self.operation.apply([tensor_value, axis_value],
                                           self.settings)

    copy_value = operation_value.copy()
    self.assertIsNot(operation_value, copy_value)
    self.assertTrue(operation_value.reconstruct_expression(use_cache=False)  # pylint: disable=g-generic-assert
                    == copy_value.reconstruct_expression(use_cache=False)
                    == 'tf.reduce_max(my_input, axis=1)')

    copy_value.operation_applications[0].arg_values[0].name = 'new_name'
    self.assertEqual(operation_value.reconstruct_expression(use_cache=False),
                     'tf.reduce_max(my_input, axis=1)')
    self.assertEqual(copy_value.reconstruct_expression(use_cache=False),
                     'tf.reduce_max(new_name, axis=1)')

  def test_reconstruct_expression(self):
    tensor_value = value.InputValue([[1, 3, 2], [-3, 0, 4]], 'my_input')
    axis_value = value.ConstantValue(1)

    operation_value = value.OperationValue(tf.constant([3, 4]),
                                           self.operation,
                                           [tensor_value, axis_value])
    expected_expression = 'tf.reduce_max(my_input, axis=1)'
    self.assertEqual(operation_value.reconstruct_expression(),
                     expected_expression)
    self.assertEqual(operation_value.reconstruct_expression(),
                     expected_expression)  # Cached.

    operation_value_apply = self.operation.apply([tensor_value, axis_value],
                                                 self.settings)
    self.assertEqual(operation_value, operation_value_apply)
    self.assertEqual(operation_value_apply.reconstruct_expression(),
                     expected_expression)

  def test_merge_reconstructions(self):
    tensor_value = value.InputValue([[1, 3, 2], [-3, 0, 4]], 'my_input')
    axis_value = value.ConstantValue(1)
    operation_value = self.operation.apply([tensor_value, axis_value],
                                           self.settings)
    self.assertLen(operation_value.operation_applications, 1)

    tensor_value_2 = value.InputValue([[1, 4], [3, 0]], 'my_input_2')
    axis_value_2 = value.ConstantValue(0)
    operation_value_2 = self.operation.apply([tensor_value_2, axis_value_2],
                                             self.settings)
    self.assertEqual(operation_value, operation_value_2)

    operation_value.merge_reconstructions(operation_value_2)
    self.assertLen(operation_value.operation_applications, 2)

  def test_reconstruct_all_expressions_with_input_names(self):
    input_0 = value.InputValue(0, 'in_0')
    constant_1 = value.ConstantValue(1)
    constant_2 = value.ConstantValue(2)

    my_input = value.InputValue([[1, 3, 2], [-3, 0, 4]], 'my_input')
    final_value = self.operation.apply([my_input, constant_1], self.settings)

    add_operation = function_operation.FunctionOperation(
        tf_functions.FunctionInfo(name='tf.add(x, y)',
                                  filter_group=filter_group.FilterGroup.NONE,
                                  weight=1))

    input_1321 = value.InputValue([[1, 3], [2, 1]], 'in_1321')
    value_2432 = add_operation.apply([input_1321, constant_1], self.settings)

    input_0210 = value.InputValue([[0, 2], [1, 0]], 'in_0210')
    value_2432_duplicate = add_operation.apply([input_0210, constant_2],
                                               self.settings)
    self.assertEqual(value_2432, value_2432_duplicate)
    value_2432.merge_reconstructions(value_2432_duplicate)

    final_value_duplicate = self.operation.apply([value_2432, input_0],
                                                 self.settings)
    self.assertEqual(final_value, final_value_duplicate)
    final_value.merge_reconstructions(final_value_duplicate)

    input_1430 = value.InputValue([[1, 4], [3, 0]], 'in_1430')
    final_value_duplicate = self.operation.apply([input_1430, input_0],
                                                 self.settings)

    self.assertEqual(final_value, final_value_duplicate)
    final_value.merge_reconstructions(final_value_duplicate)

    expected = [
        ('tf.reduce_max(my_input, axis=1)', {'my_input'}),
        ('tf.reduce_max(tf.add(in_1321, 1), axis=in_0)', {'in_1321', 'in_0'}),
        ('tf.reduce_max(tf.add(in_0210, 2), axis=in_0)', {'in_0210', 'in_0'}),
        ('tf.reduce_max(in_1430, axis=in_0)', {'in_1430', 'in_0'}),
    ]

    self.assertEqual(final_value.reconstruct_all_expressions_with_input_names(),
                     expected)

  def test_reconstruct_all_expressions_with_input_names_using_addition(self):
    constants = [value.ConstantValue(i) for i in range(10)]
    add_operation = function_operation.FunctionOperation(
        tf_functions.FunctionInfo(name='tf.add(x, y)',
                                  filter_group=filter_group.FilterGroup.NONE,
                                  weight=1))

    # The i-th element contains all unique Value objects of weight i, mapped to
    # themselves to allow retrieving the stored Value equal to some query Value.
    values_by_weight = [collections.OrderedDict()]  # Nothing of weight 0.
    # Add constants with weight 1.
    values_by_weight.append(collections.OrderedDict())
    for constant in constants:
      values_by_weight[1][constant] = constant

    for weight in range(2, 6):
      new_values = collections.OrderedDict()
      for arg_1_weight in range(1, weight):
        arg_2_weight = weight - arg_1_weight - 1
        for arg1, arg2 in itertools.product(values_by_weight[arg_1_weight],
                                            values_by_weight[arg_2_weight]):
          result = add_operation.apply([arg1, arg2], self.settings)
          if result not in new_values:
            new_values[result] = result
          else:
            new_values[result].merge_reconstructions(result)
      values_by_weight.append(new_values)

    query = value.OutputValue(9)

    # The form must be (a + b), where there are 10 choices for a, which then
    # determines b.
    reconstructions = (values_by_weight[3][query]
                       .reconstruct_all_expressions_with_input_names())
    self.assertLen(reconstructions, 10)
    # No expression uses input values.
    self.assertTrue(all(not bool(used_names)
                        for _, used_names in reconstructions))

    # No AST with only binary operators has weight 4.
    self.assertEmpty(values_by_weight[4])

    # The form is either (a + (b + c)) or ((a + b) + c). Each of the two forms
    # has 1 + 2 + ... + 9 = 45 options. Note that "a" in (a + (b + c)) cannot be
    # 0, or else (b + c) would have the same value as the entire expression.
    # Similarly, "c" in ((a + b) + c) cannot be 0.
    self.assertLen(values_by_weight[5][query]
                   .reconstruct_all_expressions_with_input_names(), 90)


class ConstantValueTest(absltest.TestCase):

  def test_init_does_not_convert_to_tensor(self):
    constant_value = value.ConstantValue(1)
    self.assertFalse(constant_value.is_tensor)
    self.assertTrue(constant_value.is_primitive)

  def test_copy(self):
    constant_value = value.ConstantValue(1)
    copy_value = constant_value.copy()
    self.assertIsNot(constant_value, copy_value)
    self.assertEqual(constant_value.reconstruct_expression(),
                     copy_value.reconstruct_expression())

  def test_reconstruct_expression(self):
    constant_value = value.ConstantValue('TF-Coder')
    self.assertEqual(constant_value.reconstruct_expression(), "'TF-Coder'")


class InputValueTest(absltest.TestCase):

  def test_init_converts_list_to_tensor(self):
    input_value = value.InputValue([1], 'descriptive_name')
    self.assertTrue(input_value.is_tensor)
    self.assertFalse(input_value.is_primitive)

  def test_init_does_not_convert_primitive_to_tensor(self):
    input_value = value.InputValue(1, 'descriptive_name')
    self.assertFalse(input_value.is_tensor)
    self.assertTrue(input_value.is_primitive)

  def test_copy(self):
    input_value = value.InputValue([1, 2], 'a')
    copy_value = input_value.copy()
    self.assertIsNot(input_value, copy_value)
    self.assertEqual(input_value.reconstruct_expression(use_cache=False),
                     copy_value.reconstruct_expression(use_cache=False))
    copy_value.name = 'b'
    self.assertNotEqual(input_value.reconstruct_expression(use_cache=False),
                        copy_value.reconstruct_expression(use_cache=False))

  def test_reconstruct_expression(self):
    input_value = value.InputValue([1], 'descriptive_name')
    self.assertEqual(input_value.reconstruct_expression(), 'descriptive_name')


class OutputValueTest(absltest.TestCase):

  def test_init_converts_to_tensor(self):
    output_value = value.OutputValue(1)
    self.assertTrue(output_value.is_tensor)

  def test_copy(self):
    with self.assertRaises(NotImplementedError):
      value.OutputValue(1).copy()

  def test_reconstruct_expression(self):
    output_value = value.OutputValue(1)
    with self.assertRaises(NotImplementedError):
      output_value.reconstruct_expression()


if __name__ == '__main__':
  logging.set_verbosity(logging.ERROR)

  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for value_search.py."""

import collections

from absl import logging
from absl.testing import absltest
from absl.testing import parameterized
import mock
import numpy as np
import six
import tensorflow as tf
from tf_coder import tf_functions
from tf_coder.benchmarks import all_benchmarks
from tf_coder.benchmarks import benchmark as benchmark_module
from tf_coder.benchmarks import simple_benchmarks
from tf_coder.models import tensor_features_model
from tf_coder.natural_language import bag_of_words_handlers
from tf_coder.natural_language import description_handler_factory
from tf_coder.value_search import all_operations
from tf_coder.value_search import value as value_module
from tf_coder.value_search import value_search
from tf_coder.value_search import value_search_settings as settings_module


def _constant(constant):
  return value_module.ConstantValue(constant)


def _inverse_sigmoid(np_array):
  return np.log(np_array / (1 - np_array))


def _parameterize_simple_benchmarks():
  """Creates parameterized test cases for all simple benchmarks.

  This is useful so that tests can run on each benchmark individually, instead
  of having a single test loop over all benchmarks. In this way, issues with
  multiple benchmarks can be identified in one round of testing, and it is
  clearer which of the benchmarks need further attention.

  Returns:
    A list of tuples (test_case_name, benchmark) for all benchmarks in
    simple_benchmarks.py.
  """
  parameterized_tuples = []
  for index, simple_benchmark in enumerate(
      all_benchmarks.all_benchmarks(modules=[simple_benchmarks])):
    # The index ensures all test cases have distinct names, even if multiple
    # benchmarks have the same name.
    test_case_name = '{index}_{name}'.format(index=index,
                                             name=simple_benchmark.name)
    parameterized_tuples.append((test_case_name, simple_benchmark))
  return parameterized_tuples


class ValueSearchTest(parameterized.TestCase):

  def setUp(self):
    super(ValueSearchTest, self).setUp()
    operations = all_operations.get_operations()
    self._constant_operation = None
    for operation in operations:
      if operation.name == tf_functions.CONSTANT_OPERATION_NAME:
        self._constant_operation = operation
        break
    self.assertIsNotNone(self._constant_operation)
    self.settings = settings_module.from_dict({'timeout': 10})

  @parameterized.named_parameters(
      ('not_sparse_list',
       [[1, 2], [3, 4, 5]],
       [1, 2, 3, 4, 5],
       'a description that does not mention that specific word',
       False),
      ('not_sparse_dict',
       {'in1': [1, 2], 'in2': [3, 4, 5]},
       [1, 2, 3, 4, 5],
       'a description that does not mention that specific word',
       False),
      ('sparse_input_list',
       [[1, 2], tf.SparseTensor(values=[1], indices=[[3]], dense_shape=[5])],
       [1, 2, 3, 4, 5],
       'a description that does not mention that specific word',
       True),
      ('sparse_input_dict',
       {'in1': [1, 2], 'in2': tf.SparseTensor(values=[1], indices=[[3]],
                                              dense_shape=[5])},
       [1, 2, 3, 4, 5],
       'a description that does not mention that specific word',
       True),
      ('sparse_output',
       [[1, 2], [3, 4, 5]],
       tf.SparseTensor(values=[1], indices=[[3]], dense_shape=[5]),
       'a description that does not mention that specific word',
       True),
      ('sparse_description',
       {'in1': [1, 2], 'in2': [3, 4, 5]},
       [1, 2, 3, 4, 5],
       'a description that does mention the word sparse',
       True),
  )
  def test_contains_sparse(self, inputs, output, description, expected):
    benchmark = benchmark_module.Benchmark(
        examples=[benchmark_module.Example(inputs, output)],
        description=description)
    self.assertEqual(value_search._contains_sparse(benchmark), expected)

  @mock.patch('sys.stdout', new_callable=six.StringIO)
  def test_add_constants_and_inputs_and_print(self, mock_stdout):
    values_by_weight = [collections.OrderedDict() for _ in range(100)]
    examples = [
        benchmark_module.Example(
            inputs=[
                [[[[1, 2, 3, 4, 5, 6, 7, 8, 9]]]],  # Shape (1, 1, 1, 9).
                [list(range(6)) for _ in range(15)],  # Shape (15, 6).
                42,
            ],
            output=tf.constant(list(range(50)), dtype=tf.int64),  # Shape (50).
        ),
    ]
    constants = [123, 45, 1, 9]
    benchmark = benchmark_module.Benchmark(examples, constants)
    output_value = value_module.OutputValue(benchmark.examples[0].output)

    value_search._add_constants_and_inputs_and_print(
        values_by_weight, benchmark, output_value, self._constant_operation,
        self.settings)

    expected_constants_ordered = [
        # Provided constants.
        _constant(123), _constant(45), _constant(1), _constant(9),
        # Input variables.
        value_module.InputValue([1], 'in1'),  # Reconstructs to 'in1'.
        value_module.InputValue([2], 'in2'),  # Reconstructs to 'in2'.
        value_module.InputValue([3], 'in3'),  # Reconstructs to 'in3'.
        # Common constants.
        _constant(0), _constant(-1), _constant(True), _constant(False),
        # Common DTypes.
        _constant(tf.int32), _constant(tf.float32), _constant(tf.bool),
        # Primitive inputs as scalar tensors. The actual Value is an
        # OperationValue, but this ExpressionValue has the same expression.
        value_module.ExpressionValue(None, 'tf.constant(in3)'),
        # Axis constants.
        _constant(2), _constant(3),
        # Uncomon DTypes.
        _constant(tf.int64),
        # Shape constants.
        _constant(6), _constant(15), _constant(50),
        # Output shape tuple.
        _constant((50,)),
    ]
    # values_by_weight is a list of OrderedDicts. Convert to list of lists.
    actual_constants_ordered = sum([list(ordered_dict)
                                    for ordered_dict in values_by_weight], [])
    self.assertEqual(
        [v.reconstruct_expression() for v in actual_constants_ordered],
        [v.reconstruct_expression() for v in expected_constants_ordered])

    # The output shape tuple (50,) is included as a constant, but not printed.
    self.assertIn(
        'Constants: [123, 45, 1, 9, 0, -1, True, False, 2, 3, 6, 15, 50]\n',
        mock_stdout.getvalue())

  @parameterized.named_parameters(
      ('keyword', 'def'),
      ('literal', 'True'),
      ('space', ' '),
      ('number', '0'),
      ('tf', 'tf'),
      ('np', 'np'),
      ('dot', 'x.y'),
      ('not_string', 1))
  def test_add_constants_and_inputs_and_print_checks_names(self, bad_name):
    values_by_weight = [collections.OrderedDict() for _ in range(100)]
    examples = [benchmark_module.Example(inputs={bad_name: [1, 2, 3]},
                                         output=[1, 2])]
    benchmark = benchmark_module.Benchmark(examples, constants=[])
    output_value = value_module.OutputValue(benchmark.examples[0].output)
    with self.assertRaises(ValueError):
      value_search._add_constants_and_inputs_and_print(
          values_by_weight, benchmark, output_value, self._constant_operation,
          self.settings)

  @parameterized.named_parameters(
      ('with_description', 'do foo to tensor', True),
      ('empty_description', '', False),
      ('none_description', None, False))
  @mock.patch('sys.stdout', new_callable=six.StringIO)
  def test_add_constants_and_inputs_and_print_handles_description(
      self, description, expect_print, mock_stdout):
    examples = [benchmark_module.Example(inputs=[[1, 4], [2, 7]],
                                         output=[3, 11])]
    benchmark = benchmark_module.Benchmark(examples, description=description)
    values_by_weight = [collections.OrderedDict() for _ in range(5)]
    output_value = value_module.OutputValue(benchmark.examples[0].output)
    value_search._add_constants_and_inputs_and_print(
        values_by_weight, benchmark, output_value, self._constant_operation,
        self.settings)
    if expect_print:
      self.assertIn('Description: {}\n'.format(description),
                    mock_stdout.getvalue())
    else:
      self.assertNotIn('Description:', mock_stdout.getvalue())

  @parameterized.named_parameters(_parameterize_simple_benchmarks())
  def test_run_value_search_works_for_simple_benchmarks(self, simple_benchmark):
    handler = description_handler_factory.create_handler('tfidf')
    results = value_search.run_value_search(
        benchmark=simple_benchmark,
        settings=settings_module.from_dict({
            'timeout': 20,
            'printing.statistics': True,
            'require_all_inputs_used': False,
            'require_one_input_used': False,
        }),
        description_handler=handler)
    self.assertLen(results.solutions, 1)
    self.assertIsNotNone(results.solutions[0].expression)
    self.assertGreater(results.solutions[0].weight, 0)
    self.assertGreater(results.total_time, 0)
    self.assertNotEmpty(results.value_set)
    self.assertGreaterEqual(results.statistics.total_apply_count, 0)
    # The simple benchmarks are so simple that we can directly write down the
    # correct (minimal) target program.
    self.assertEqual(results.solutions[0].expression,
                     simple_benchmark.target_program)

  @parameterized.named_parameters(
      ('stackoverflow_06', 'stackoverflow_06', 10.0,
       'tf.cast(tf.equal(in1, tf.expand_dims(in1, 1)), tf.float32)'),
      ('stackoverflow_22', 'stackoverflow_22', 10.0,
       'tf.tensordot(tf.cast(in1, tf.float32), in2, 1)'),
      ('google_01', 'google_01', 90.0,
       'tf.cast(tf.where(tf.sequence_mask(tf.math.bincount(in1))), tf.int32)'))
  def test_run_value_search_works_for_important_benchmarks(
      self, benchmark_name, timeout, solution):
    # These benchmarks are frequently used in demos. We should ensure that they
    # are solved quickly and consistently. While the exact solution may change
    # as the algorithm improves, we should ensure that the solution is always
    # simple and understandable.
    handler = description_handler_factory.create_handler('tfidf')

    benchmark = all_benchmarks.find_benchmark_with_name(benchmark_name)
    results = value_search.run_value_search(
        benchmark=benchmark,
        description_handler=handler,
        settings=settings_module.from_dict({'timeout': timeout}))
    print('Time for benchmark {}: {:.2f} sec'.format(
        benchmark_name, results.total_time))
    self.assertLen(results.solutions, 1)
    self.assertEqual(results.solutions[0].expression, solution)

  @parameterized.named_parameters(
      ('true', True, ['tf.add(in1, in2)', 'tf.add(in2, in1)']),
      ('false', False, ['tf.add(in1, in2)',
                        'tf.add(in2, in1)',
                        'tf.add_n((in1, in2))',
                        'tf.add_n((in2, in1))']))
  def test_run_value_search_only_minimal_solutions(
      self, only_minimal_solutions, expected_solutions):
    benchmark = benchmark_module.Benchmark(
        examples=[benchmark_module.Example(inputs=[[1, 4], [2, 7]],
                                           output=[3, 11])])
    results = value_search.run_value_search(
        benchmark=benchmark,
        settings=settings_module.from_dict({
            'timeout': 20,
            'max_solutions': 4,
            'only_minimal_solutions': only_minimal_solutions,
            'max_extra_solutions_time': 20}))
    self.assertLen(results.solutions, len(expected_solutions))
    self.assertEqual([solution.expression for solution in results.solutions],
                     expected_solutions)

  @parameterized.named_parameters(
      ('stackoverflow_06', 'stackoverflow_06', 0),
      ('stackoverflow_22', 'stackoverflow_22', 1))
  def test_run_value_search_works_for_small_max_weight(
      self, benchmark_name, max_weight):
    benchmark = all_benchmarks.find_benchmark_with_name(benchmark_name)
    results = value_search.run_value_search(
        benchmark=benchmark,
        settings=settings_module.from_dict({'max_weight': max_weight}))
    print("Note: if you're inspecting the test log, the above failure is "
          "expected in test_run_value_search_works_for_small_max_weight.")
    self.assertEmpty(results.solutions)

  def test_run_value_handles_large_weight_constants(self):
    benchmark = benchmark_module.Benchmark(
        examples=[benchmark_module.Example(inputs=[[1], [2]], output=[[3]])])
    results = value_search.run_value_search(
        benchmark=benchmark, settings=self.settings)
    self.assertNotEmpty(results.solutions)
    self.assertEqual(results.solutions[0].expression,
                     'tf.add(in1, tf.expand_dims(in2, 0))')
    output_shape_constant = value_module.ConstantValue((1, 1))
    self.assertIn(output_shape_constant, results.value_set)
    # Find the element in value_set equal to output_shape_constant and assert
    # that it's actually a ConstantValue, as opposed to an OperationValue.
    for value in results.value_set:
      if value == output_shape_constant:
        self.assertIsInstance(value, value_module.ConstantValue)

  def test_run_value_search_from_example(self):
    results = value_search.run_value_search_from_example(
        inputs=[
            [1, 2, 3],
            [10, 20, 30]
        ],
        output=[11, 22, 33],
        settings=settings_module.from_dict({
            'timeout': 5, 'max_solutions': 2}))
    self.assertLen(results.solutions, 2)
    self.assertEqual(results.solutions[0].expression, 'tf.add(in1, in2)')
    self.assertEqual(results.solutions[1].expression, 'tf.add(in2, in1)')
    # Linter suggests self.assertLess() but it's wrong.
    self.assertTrue(0.0 < results.total_time < 5.0)  # pylint: disable=g-generic-assert

  def test_run_value_search_from_example_with_constants(self):
    results = value_search.run_value_search_from_example(
        inputs=[
            [10, 20, 30],
        ],
        output=[54, 64, 74],
        settings=self.settings,
        constants=[44])
    self.assertLen(results.solutions, 1)
    self.assertEqual(results.solutions[0].expression,
                     'tf.add(in1, tf.constant(44))')

  def test_value_search_checks_bad_solutions(self):
    inputs = {'tensor': [10, 20, 30],
              'index': -2}
    output = 20
    constants = [1, 20]

    results = value_search.run_value_search_from_example(
        inputs=inputs, output=output, constants=constants,
        settings=settings_module.from_dict({'require_all_inputs_used': True}))
    self.assertLen(results.solutions, 1)
    self.assertEqual(results.solutions[0].expression, 'tensor[index]')

    results = value_search.run_value_search_from_example(
        inputs=inputs, output=output, constants=constants,
        settings=settings_module.from_dict({'require_all_inputs_used': False,
                                            'require_one_input_used': True}))
    self.assertLen(results.solutions, 1)
    self.assertEqual(results.solutions[0].expression, 'tensor[1]')

    results = value_search.run_value_search_from_example(
        inputs=inputs, output=output, constants=constants,
        settings=settings_module.from_dict({'require_all_inputs_used': False,
                                            'require_one_input_used': False}))
    self.assertLen(results.solutions, 1)
    self.assertEqual(results.solutions[0].expression, 'tf.constant(20)')

  def test_value_search_prioritizes_operations(self):
    # Make sure prioritized ops get printed out.
    settings = settings_module.from_dict({
        'timeout': 0.5,  # We don't actually care about solving the problem.
        'printing.prioritized_operations': True,
        'printing.deprioritized_operations': True,
        'tensor_model.prioritize_threshold': 0.5,
        'tensor_model.deprioritize_threshold': 0.2,
    })

    # Create data for the mocks.
    all_ops = all_operations.get_operations(include_sparse_operations=True)
    op_names = [op.name for op in all_ops]
    abs_index = None  # Prioritized by tensor features model.
    argmax_index = None  # Deprioritized by tensor features model.
    for i, op in enumerate(all_ops):
      if op.name == 'tf.abs(x)':
        abs_index = i
      elif op.name == 'tf.argmax(input, axis)':
        argmax_index = i
    self.assertIsNotNone(abs_index)
    self.assertIsNotNone(argmax_index)

    operation_probs = np.repeat(0.4, len(all_ops))
    operation_probs[abs_index] = 0.6  # Above prioritize threshold.
    operation_probs[argmax_index] = 0.1  # Below deprioritize threshold.
    operation_logits = np.expand_dims(_inverse_sigmoid(operation_probs), axis=0)

    nl_data = [
        # Associate "apple" with tf.zeros.
        {'docstring': ['apple pie'], 'tf_functions': ['tf.zeros'],
         'comments': [], 'names': [], 'strings': []},
    ] * 100

    # Mock the tensor model's predictions, the NL model's data, and print().
    with mock.patch('tf_coder.models.tensor_features_model.eval_single_example',
                    return_value=tensor_features_model.Result(
                        operation_logits=operation_logits)), \
        mock.patch('tf_coder.datasets.github.data_loader.load_data',
                   return_value=nl_data), \
        mock.patch('sys.stdout', new_callable=six.StringIO) as mock_stdout:

      handler = bag_of_words_handlers.NaiveBayesDescriptionHandler(
          max_num_prioritized=1)
      benchmark = benchmark_module.Benchmark(
          # I/O example doesn't matter.
          examples=[benchmark_module.Example({'my_var': [1, 2]}, [2, 1])],
          # Description contains "apple"!
          description='honeycrisp apple')

      value_search.run_value_search(
          benchmark=benchmark,
          description_handler=handler,
          settings=settings,
          tensor_model=mock.Mock(),
          tensor_config={'max_num_inputs': 3,
                         'operation_names': op_names})

      self.assertIn('BOW handler prioritized tf.zeros',
                    mock_stdout.getvalue())
      self.assertIn('Tensor features model prioritized tf.abs(x), p=0.6',
                    mock_stdout.getvalue())
      self.assertIn(
          'Tensor features model deprioritized tf.argmax(input, axis), p=0.1',
          mock_stdout.getvalue())

  @mock.patch('sys.stdout', new_callable=six.StringIO)
  def test_value_search_can_load_data(self, mock_stdout):
    benchmark = all_benchmarks.find_benchmark_with_name('simple_cast')
    handler = description_handler_factory.create_handler('naive_bayes')

    settings = settings_module.from_dict({
        'timeout': 5,
        'printing.prioritized_operations': True,
        'printing.deprioritized_operations': True,
    })

    tensor_config = tensor_features_model.load_config(
        settings.tensor_model.config_path)
    tensor_model = tensor_features_model.get_model(tensor_config)
    tensor_checkpoint = tensor_features_model.create_checkpoint(tensor_model)
    tensor_checkpoint.restore(
        settings.tensor_model.checkpoint_path).expect_partial()

    results = value_search.run_value_search(
        benchmark=benchmark,
        description_handler=handler,
        settings=settings,
        tensor_model=tensor_model,
        tensor_config=tensor_config)

    self.assertLen(results.solutions, 1)
    self.assertIn('BOW handler prioritized tf.cast(x, dtype)',
                  mock_stdout.getvalue())
    self.assertIn('Tensor features model prioritized tf.cast(x, dtype)',
                  mock_stdout.getvalue())


if __name__ == '__main__':
  logging.set_verbosity(logging.ERROR)

  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Installs TF-Coder."""

import setuptools

from tf_coder import __version__


def get_required_packages():
  """Returns a list of required packages."""
  return [
      'absl-py >= 0.6.1',  # Oct 26, 2018
      'funcsigs >= 1.0.2',  # Apr 25, 2016
      'numpy >= 1.15.4',  # Nov 4, 2018
      'scikit-learn >= 0.22.2',  # Feb 28, 2020
      'six >= 1.12.0',  # Dec 9, 2018
      'tensorflow >= 2.2.0',  # May 7, 2020
  ]


def run_setup():
  """Installs TF-Coder."""

  with open('README.md', 'r') as fh:
    long_description = fh.read()

  setuptools.setup(
      name='tensorflow-coder',
      version=__version__,
      author='Google LLC',
      author_email='no-reply@google.com',
      description=('TensorFlow Coder (TF-Coder): '
                   'A Program Synthesis Tool for TensorFlow'),
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/google-research/tensorflow-coder',
      packages=setuptools.find_packages(
          exclude=['tf_coder_colab_logging', 'tf_coder_colab_logging.*']),
      package_data={
          'tf_coder.datasets.github': [
              'data/tf_apis_full.tfrecord-00000-of-00001',
          ],
          'tf_coder.models': [
              'trained_model/ckpt-1172.data-00000-of-00001',
              'trained_model/config.json',
              'trained_model/ckpt-1172.index',
          ],
      },
      include_package_data=True,
      install_requires=get_required_packages(),
      classifiers=[
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Topic :: Scientific/Engineering :: Artificial Intelligence',
          'Topic :: Software Development',
          'Topic :: Software Development :: Code Generators',
      ],
      keywords=('tensorflow coder tf-coder program synthesis tensor '
                'manipulation programming by example'),
      python_requires='>=3',
  )


if __name__ == '__main__':
  run_setup()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Computes the size of value search's search space."""

import collections
import functools
import operator
import os
import sys

from absl import app
from absl import flags
from tf_coder import tf_coder_utils
from tf_coder import tf_functions
from tf_coder.benchmarks import all_benchmarks
from tf_coder.natural_language import description_handler_factory
from tf_coder.value_search import value as value_module
from tf_coder.value_search import value_search
from tf_coder.value_search import value_search_settings as settings_module


FLAGS = flags.FLAGS

flags.DEFINE_string('benchmark_name', 'google_02',
                    'The name of a benchmark to analyze.')
flags.DEFINE_multi_string('settings',
                          [],
                          'Settings to override the defaults.')


# Inspired by https://stackoverflow.com/a/45669280/9589593.
class SuppressPrint(object):
  """A context manager for suppressing print() calls temporarily."""

  def __enter__(self):
    self._old_stdout = sys.stdout
    sys.stdout = open(os.devnull, 'w')

  def __exit__(self, exc_type, exc_val, exc_tb):
    del exc_type, exc_val, exc_tb
    sys.stdout.close()
    sys.stdout = self._old_stdout


def compute_search_space_size(benchmark, settings, description_handler):
  """Computes and prints the size of the search space.

  This counts the total number of expressions with weight at most max_weight.
  The weights come from the benchmark (for constants and inputs) and the
  description handler (for determining the op weights). Distinct expressions
  will be counted separately even if they evaluate to the same value, unlike in
  TF-Coder's value_search algorithm which does value-based pruning.

  Args:
    benchmark: The Benchmark object defining the problem to analyze.
    settings: A Settings object containing settings for value search.
    description_handler: The DescriptionHandler used, which can modify weights
      of operations.

  Returns:
    Nothing. All output is printed to stdout.
  """

  max_weight = settings.max_weight
  print('Computing search space.\n'
        'Benchmark name: {}\n'
        'Description handler: {}\n'
        'Max weight: {}'.format(
            benchmark.name, description_handler, max_weight))

  # TODO(kshi): Update to load the tensor features model/config.
  operations = value_search.get_reweighted_operations(benchmark,
                                                      settings,
                                                      description_handler,
                                                      tensor_model=None,
                                                      tensor_config=None)

  # These loops are not the most efficient, but it doesn't really matter.
  print('\nFound {} operations.'.format(len(operations)))
  print()
  for weight in range(1, max(op.weight for op in operations) + 1):
    print('# operations with weight {}: {}'.format(
        weight, sum(1 for op in operations if op.weight == weight)))
  print()
  for arity in range(1, max(op.num_args for op in operations) + 1):
    print('# operations with arity {}: {}'.format(
        arity, sum(1 for op in operations if op.num_args == arity)))

  output_value = value_module.OutputValue(benchmark.examples[0].output)
  values_by_weight = [collections.OrderedDict()
                      for _ in range(max_weight + 1)]

  constant_operation = None
  for operation in operations:
    if operation.name == tf_functions.CONSTANT_OPERATION_NAME:
      constant_operation = operation
      break
  with SuppressPrint():
    value_search._add_constants_and_inputs_and_print(  # pylint: disable=protected-access
        values_by_weight, benchmark, output_value, constant_operation, settings)

  num_expressions_with_weight = [len(values_with_weight)
                                 for values_with_weight in values_by_weight]
  print()
  max_weight_with_initial_value = max(w for w in range(max_weight + 1)
                                      if num_expressions_with_weight[w])
  for weight in range(1, max_weight_with_initial_value + 1):
    print('# initial values with weight {}: {}'.format(
        weight, num_expressions_with_weight[weight]))

  for total_weight in range(2, max_weight + 1):
    for operation in operations:
      # All operations should have strictly positive weight and num_args.
      op_weight = operation.weight
      op_arity = operation.num_args

      if total_weight - op_weight < op_arity:
        continue

      # Partition `total_weight - op_weight` into `op_arity` positive pieces.
      # Equivalently, partition `total_weight - op_weight - op_arity` into
      # `op_arity` nonnegative pieces.
      for partition in tf_coder_utils.generate_partitions(
          total_weight - op_weight - op_arity, op_arity):
        arg_weights = [part + 1 for part in partition]
        num_expressions_with_weight[total_weight] += functools.reduce(
            operator.mul,
            (num_expressions_with_weight[w] for w in arg_weights))

  print()
  for weight in range(1, max_weight + 1):
    print('# expressions with weight exactly {}: {}'.format(
        weight, num_expressions_with_weight[weight]))

  print()
  for weight in range(1, max_weight + 1):
    print('# expressions with weight up to {}: {}'.format(
        weight, sum(num_expressions_with_weight[:weight + 1])))


def main(unused_argv):
  settings = settings_module.from_list(FLAGS.settings)
  description_handler = description_handler_factory.create_handler(
      settings.description_handler_name)
  benchmark = all_benchmarks.find_benchmark_with_name(FLAGS.benchmark_name)
  if not benchmark:
    raise ValueError('Unknown benchmark: {}'.format(FLAGS.benchmark_name))

  compute_search_space_size(benchmark=benchmark,
                            settings=settings,
                            description_handler=description_handler)

if __name__ == '__main__':
  app.run(main)



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for random_inputs.py."""

from absl.testing import absltest
from tf_coder.datasets import random_inputs
from tf_coder.value_search import value as value_module


class SeedInputsTest(absltest.TestCase):

  def test_generate_random_input(self):
    for i in range(1000):
      # Just check that generation completes and produces valid Values.
      random_input = random_inputs.generate_random_input(random_seed=i)
      value = value_module.InputValue(random_input, name='dummy_name')
      self.assertIsNotNone(value)

  def test_generate_random_input_uses_seed(self):
    random_input_1a = random_inputs.generate_random_input(random_seed=1)
    random_input_1b = random_inputs.generate_random_input(random_seed=1)
    random_input_2 = random_inputs.generate_random_input(random_seed=2)

    self.assertEqual(str(random_input_1a), str(random_input_1b))
    self.assertNotEqual(str(random_input_1a), str(random_input_2))


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Manages all Operation objects used by value search."""

import inspect
from typing import List, Optional, Text

from tf_coder import tf_functions
from tf_coder.value_search import function_operation
from tf_coder.value_search import operation_base
from tf_coder.value_search import python_operations


def get_python_operations() -> List[operation_base.Operation]:
  """Returns a list of Operation objects from the python_operations module."""
  operation_classes = inspect.getmembers(
      python_operations,
      lambda x: inspect.isclass(x) and not inspect.isabstract(x))
  return [operation_class()
          for unused_name, operation_class in operation_classes]


def get_tf_operations() -> List[operation_base.Operation]:
  """Returns a list of Operation objects for dense TensorFlow operations."""
  return [function_operation.FunctionOperation(function_info)
          for function_info in tf_functions.TF_FUNCTIONS]


def get_sparse_operations() -> List[operation_base.Operation]:
  """Returns a list of Operation objects for sparse operations."""
  return [function_operation.FunctionOperation(function_info)
          for function_info in tf_functions.SPARSE_FUNCTIONS]


def get_operations(
    include_sparse_operations: bool = False) -> List[operation_base.Operation]:
  """Returns a list of Operation objects that value search should use."""
  operations = []
  operations.extend(get_tf_operations())
  if include_sparse_operations:
    operations.extend(get_sparse_operations())
  operations.extend(get_python_operations())
  return operations


def find_operation_with_name(
    operation_name: Text,
    operation_list: Optional[List[operation_base.Operation]] = None
) -> operation_base.Operation:
  """Finds an operation with the given name, optionally within a given list."""
  if operation_list is None:
    operation_list = get_operations(include_sparse_operations=True)
  matching_operations = [op for op in operation_list
                         if op.name == operation_name]
  if len(matching_operations) == 1:
    return matching_operations[0]

  raise ValueError('Found {} operations matching the name {}'.format(
      len(matching_operations), operation_name))



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Benchmarks collected/inspired from StackOverflow."""

# Avoid wrapping URLs and target programs to ease clicking and copying.
# pylint: disable=line-too-long

# Every function in this module takes no arguments and creates a benchmark.
# pylint: disable=missing-docstring

import math

import tensorflow as tf
from tf_coder.benchmarks import benchmark


def stackoverflow_01():
  examples = [
      benchmark.Example(
          inputs=[
              [[5., 2.], [1., 3.], [0., -1.]],
          ],
          output=[[[5., 5.], [1., 1.], [0., 0.]],
                  [[2., 2.], [3., 3.], [-1., -1.]]]
      ),
  ]
  constants = []
  description = 'reshape by separating and duplicating columns'
  target_program = 'tf.transpose(tf.cast((in1, in1), tf.float32))'
  source = 'https://stackoverflow.com/questions/40441503/tensorflow-tensor-reshape'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_01')


def stackoverflow_02():
  examples = [
      benchmark.Example(
          inputs=[
              [5, 1, 0, 3, 0, -1, 2, -10, 2],
          ],
          output=[1, 1, 0, 1, 0, -1, 1, -10, 1]
      ),
  ]
  constants = [1]
  description = 'clip values that are greater than 1'
  target_program = 'tf.minimum(in1, tf.constant(1))'
  source = 'https://stackoverflow.com/questions/46408839/tensorflow-trim-values-in-tensor'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_02')


def stackoverflow_03():
  examples = [
      benchmark.Example(
          inputs=[
              [[11, 22, 33, 44, 55, 66, 77], [70, 60, 50, 40, 30, 20, 10]],
              [[-9, -8, -7, -6, -5, -4, -3], [11, 12, 13, 14, 15, 16, 17]],
          ],
          output=[[11, 22, 33, -6, -5, 66, 77], [70, 60, 50, 14, 15, 20, 10]]
      ),
  ]
  constants = [3, 4, 5]
  description = 'replace certain columns with columns from the other tensor'
  # Solution is simpler when using auxiliary variables:
  # mask = tf.reduce_sum(tf.one_hot(tf.range(3, 5), depth=in1.shape[1],
  #                                 dtype=tf.int32), axis=0)
  # output = mask * in2 + (1 - mask) * in1
  target_program = 'tf.reduce_sum(tf.one_hot(tf.range(3, 5), depth=in1.shape[1], dtype=tf.int32), axis=0) * in2 + (1 - tf.reduce_sum(tf.one_hot(tf.range(3, 5), depth=in1.shape[1], dtype=tf.int32), axis=0)) * in1'
  source = 'https://stackoverflow.com/questions/44657388/how-to-replace-certain-values-in-tensorflow-tensor-with-the-values-of-the-other'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_03')


def stackoverflow_04():
  examples = [
      benchmark.Example(
          inputs=[
              [[12, 23, 34, 45], [66, 77, 88, 99]],
              [[0, 1], [0, 1], [1, 0], [0, 0]],
              [[2, 1], [1, 2], [0, 2], [0, 0]],
          ],
          output=[[34, 77], [23, 88], [66, 34], [12, 12]]
      ),
  ]
  constants = []
  description = 'index into the tensor'
  target_program = 'tf.gather_nd(in1, tf.stack((in2, in3), axis=-1))'
  source = 'https://stackoverflow.com/questions/33736795/tensorflow-numpy-like-tensor-indexing'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_04')


def stackoverflow_05():
  examples = [
      benchmark.Example(
          inputs=[
              [[4, 3, 1], [6, 5, 2]],
              [[[5, 5]], [[1, 5]], [[6, 0]]],
          ],
          output=[[[29, 35]], [[47, 55]]]
      ),
  ]
  constants = []
  description = 'tensor multiplication like np.tensordot'
  target_program = 'tf.tensordot(in1, in2, 1)'
  source = 'https://stackoverflow.com/questions/43067338/tensor-multiplication-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_05')


def stackoverflow_06():
  examples = [
      benchmark.Example(
          inputs=[
              [3, 5, 0, 2, 3, 3, 0],
          ],
          output=[[1., 0., 0., 0., 1., 1., 0.],
                  [0., 1., 0., 0., 0., 0., 0.],
                  [0., 0., 1., 0., 0., 0., 1.],
                  [0., 0., 0., 1., 0., 0., 0.],
                  [1., 0., 0., 0., 1., 1., 0.],
                  [1., 0., 0., 0., 1., 1., 0.],
                  [0., 0., 1., 0., 0., 0., 1.]]
      ),
  ]
  constants = []
  description = 'binary tensor from vector indicating if elements are equal'
  target_program = 'tf.cast(tf.equal(in1, tf.expand_dims(in1, 1)), tf.float32)'
  source = 'https://stackoverflow.com/questions/47816231/create-binary-tensor-from-vector-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_06')


def stackoverflow_07():
  examples = [
      benchmark.Example(
          inputs=[
              [[[8, 4, 6], [2, 12, 3]],
               [[11, 12, 5], [9, 12, 12]],
               [[9, 2, 13], [7, 0, 7]],
               [[2, 10, 5], [7, 1, 2]]],
          ],
          output=[[[8, 4, 6], [11, 12, 5], [9, 2, 13], [2, 10, 5]],
                  [[2, 12, 3], [9, 12, 12], [7, 0, 7], [7, 1, 2]]]
      ),
  ]
  constants = []
  description = 'swap the first two dimensions of the tensor'
  target_program = 'tf.cast(tf.unstack(in1, axis=1), tf.int32)'
  source = 'https://stackoverflow.com/questions/38212205/swap-tensor-axes-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_07')


def stackoverflow_08():
  examples = [
      benchmark.Example(
          inputs=[
              [-1, 0, -3, 2, 1, 3, 5, -1, -9, 2, 10],
              [12, 3, 45, 6, 7, 8, 9, 87, 65, 4, 32],
          ],
          output=[6, 8, 9, 4, 32]
      ),
  ]
  constants = [1]
  description = ('select the values in the second tensor where the first '
                 'tensor is greater than 1')
  target_program = 'tf.boolean_mask(in2, tf.greater(in1, tf.constant(1)))'
  source = 'https://stackoverflow.com/questions/33769041/tensorflow-indexing-with-boolean-tensor'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_08')


def stackoverflow_09():
  examples = [
      benchmark.Example(
          inputs=[
              [37, 42, 42, 37, 28, 15, 42, 15],
          ],
          output=[0, 1, 1, 0, 2, 3, 1, 3]
      ),
  ]
  constants = []
  description = 'group items by value and get the group indices'
  target_program = 'tf.unique_with_counts(in1)[1]'
  source = 'https://stackoverflow.com/questions/53054668/assign-values-between-0-and-n-1-for-a-vector-of-length-l-with-n-different-eleme'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_09')


def stackoverflow_10():
  examples = [
      benchmark.Example(
          inputs=[
              [[15, 10], [20, -5]],
              [[2, 3, 1], [-2, 5, 0]],
          ],
          output=[[[30, 45, 15], [20, 30, 10]],
                  [[-40, 100, 0], [10, -25, 0]]]
      ),
  ]
  constants = []
  description = 'perform matrix multiplication'
  target_program = 'tf.matmul(tf.expand_dims(in1, -1), tf.expand_dims(in2, 1))'
  source = 'https://stackoverflow.com/questions/53094212/tensorflow-sxn-matrix-multiply-with-sxd-matrix-to-output-sxnxd-array'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_10')


def stackoverflow_11():
  examples = [
      benchmark.Example(
          inputs=[
              [4, 0, 1, 1, 0, 4, 0, 0, 3, 4, 1],
          ],
          output=[4, 3, 0, 1, 3]
      ),
  ]
  constants = []
  description = 'count the number of occurences of each distinct number'
  target_program = 'tf.math.bincount(in1)'
  source = 'https://stackoverflow.com/questions/45194672/how-to-count-elements-in-tensorflow-tensor'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_11')


def stackoverflow_12():
  examples = [
      benchmark.Example(
          inputs=[
              [[12, 34, 56], [33, 22, 11]]
          ],
          output=[[12, 56], [33, 11]]
      ),
  ]
  constants = [0, 1, 2]
  description = 'remove a column from the tensor'
  target_program = 'tf.gather(in1, (0, 2), axis=1, batch_dims=0)'
  source = 'https://stackoverflow.com/questions/47447183/remove-a-set-of-tensors-from-a-tensor-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_12')


def stackoverflow_13():
  examples = [
      benchmark.Example(
          inputs=[
              [[3, 5], [10, 2]],
              [[[1, 0], [5, 4]], [[3, 10], [2, -2]]],
          ],
          output=[[[28, 20], [19, 20]], [[20, 8], [34, 96]]]
      ),
  ]
  constants = []
  description = 'multiply vectors by tensor'
  target_program = 'tf.tensordot(in1, in2, (1, 1))'
  source = 'https://stackoverflow.com/questions/50777704/n-d-tensor-matrix-multiplication-with-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_13')


def stackoverflow_14():
  examples = [
      benchmark.Example(
          inputs=[
              [[[False, False, True],
                [False, False, False],
                [True, False, True],
                [False, True, False],
                [False, False, False],
                [True, True, True],
                [True, True, False]]],
          ],
          output=[[True, False, True, True, False, True, True]]
      ),
  ]
  constants = []
  target_program = 'tf.reduce_any(in1, axis=-1)'
  description = 'choose True if any value in a row is True, False otherwise'
  source = 'https://stackoverflow.com/questions/35657003/aggregate-each-element-of-tensor-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_14')


def stackoverflow_15():
  examples = [
      benchmark.Example(
          inputs=[
              [3, 1, 2, 0, 1, -1, 10, 1, -10],
          ],
          output=[3, 0, 2, 0, 0, -1, 10, 0, -10]
      ),
  ]
  constants = [0, 1]
  description = 'set all instances of 1 to 0'
  target_program = 'tf.subtract(in1, tf.cast(tf.equal(in1, tf.constant(1)), tf.int32))'
  source = 'https://stackoverflow.com/questions/39045797/conditional-assignment-of-tensor-values-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_15')


def stackoverflow_16():
  examples = [
      benchmark.Example(
          inputs=[
              [[2, 5], [3, 0], [8, -7]],
              [4, 10, -6],
          ],
          output=[[8, 20], [30, 0], [-48, 42]]
      ),
  ]
  constants = []
  description = 'multiply tensors across the first axis'
  target_program = 'tf.multiply(in1, tf.expand_dims(in2, 1))'
  source = 'https://stackoverflow.com/questions/46240646/tensor-multiply-along-axis-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_16')


def stackoverflow_17():
  examples = [
      benchmark.Example(
          inputs=[
              [17, -32, 99],
          ],
          output=[[17, 17], [-32, -32], [99, 99]]
      ),
  ]
  constants = []
  description = 'duplicate each element of a tensor'
  # StackOverflow answer doesn't work.
  target_program = 'tf.stack((in1, in1), axis=1)'
  source = 'https://stackoverflow.com/questions/51761353/about-tensor-of-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_17')


def stackoverflow_18():
  examples = [
      benchmark.Example(
          inputs=[
              # shape=[2, 2, 3].
              [[[1, 1, 1], [1, 0, 1]],
               [[1, 2, 3], [4, 5, 6]]],
              # shape=[3, 4].
              [[1, 1, 1, 1], [1, 2, 3, 4], [5, 6, 7, 8]],
              # shape=[4].
              [100, 200, 300, 400],
          ],
          # Shape=[sequence_length, batch_size, 4]=[2, 2, 4].
          output=[[[107, 209, 311, 413], [106, 207, 308, 409]],
                  [[118, 223, 328, 433], [139, 250, 361, 472]]]
      ),
  ]
  constants = []
  description = 'multiply 3D tensor and 2D tensor and add another tensor'
  target_program = 'tf.add(in3, tf.matmul(in1, in2))'
  source = 'https://stackoverflow.com/questions/38222126/tensorflow-efficient-way-for-tensor-multiplication'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_18')


def stackoverflow_19():
  examples = [
      benchmark.Example(
          inputs=[
              [[3, 1, 2],
               [1, 0, 4],
               [1, 2, 3],
               [0, 5, 1],
               [1, 1, 2],
               [2, 3, 1],
               [2, 1, 0]],
          ],
          output=[[0, 5, 1],
                  [1, 0, 4],
                  [1, 1, 2],
                  [1, 2, 3],
                  [2, 1, 0],
                  [2, 3, 1],
                  [3, 1, 2]]
      ),
  ]
  constants = []
  description = ('sort a tensor considering the first column, breaking ties '
                 'using the second column')
  # Needs new variable to avoid code duplication:
  # temp = tf.gather(in1, tf.argsort(in1[:, 1], stable=True))
  # output = tf.gather(temp, tf.argsort(temp[:, 0], stable=True))
  target_program = 'tf.gather(tf.gather(in1, tf.argsort(in1[:, 1], stable=True)), tf.argsort(tf.gather(in1, tf.argsort(in1[:, 1], stable=True))[:, 0], stable=True))'
  source = 'https://stackoverflow.com/questions/49399198/sort-a-tensor-based-on-two-columns-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_19')


def stackoverflow_20():
  examples = [
      benchmark.Example(
          inputs=[
              [[0.7, 0.2, 0.1],
               [0.4, 0.5, 0.1],
               [0.4, 0.4, 0.2],
               [0.3, 0.4, 0.3],
               [0.0, 0.0, 1.0]],
          ],
          output=[[1, 0, 0],
                  [0, 1, 0],
                  [1, 0, 0],
                  [0, 1, 0],
                  [0, 0, 1]]
      ),
  ]
  constants = []
  description = 'compute argmax in each tensor and set it to 1'
  target_program = 'tf.cast(tf.one_hot(tf.argmax(in1, axis=1), 3), tf.int32)'
  source = 'https://stackoverflow.com/questions/44834739/argmax-on-a-tensor-and-ceiling-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_20')


def stackoverflow_21():
  examples = [
      benchmark.Example(
          inputs=[
              [[2], [0], [1], [0]],
              [[0.2, 0.5, 0.3],
               [0.1, 0.3, 0.6],
               [0.1, 0.6, 0.3],
               [0.7, 0.0, 0.3]],
          ],
          output=[[0.3], [0.1], [0.6], [0.7]]
      ),
  ]
  constants = []
  description = 'gather elements in a tensor along axis 1'
  target_program = 'tf.gather(in2, in1, axis=1, batch_dims=1)'
  source = 'https://stackoverflow.com/questions/51690095/how-to-gather-element-with-index-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_21')


def stackoverflow_22():
  examples = [
      benchmark.Example(
          inputs=[
              [3, 1, 10],
              [[0.6, 0.4], [0.5, 1.0], [3.0, 4.0]],
          ],
          output=[32.3, 42.2]
      ),
  ]
  constants = []
  description = 'multiply a vector with a matrix without reshaping the vector'
  target_program = 'tf.tensordot(tf.cast(in1, tf.float32), in2, 1)'
  source = 'https://stackoverflow.com/questions/43284897/how-can-i-multiply-a-vector-and-a-matrix-in-tensorflow-without-reshaping'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_22')


def stackoverflow_23():
  # Simplified slightly because the user already knows how to do the mod part.
  examples = [
      benchmark.Example(
          inputs=[
              [[0, 5, 2], [3, 1, 4], [5, 1, 5]],
          ],
          output=[[1, 0, 1, 0, 0, 1, 0, 0, 0],
                  [0, 1, 0, 1, 1, 0, 0, 0, 0],
                  [0, 1, 0, 0, 0, 1, 0, 0, 0]]
      ),
  ]
  constants = []
  description = 'place 1 at the indices in the input tensor'
  target_program = 'tf.cast(tf.reduce_max(tf.one_hot(in1, 9), axis=1), tf.int32)'
  source = 'https://stackoverflow.com/questions/53414433/tensorflow-tensor-binarization'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_23')


def stackoverflow_24():
  examples = [
      benchmark.Example(
          inputs=[
              [3.0, 1.0, 4.0, 5.0, 2.0, 8.0, -6.0, -7.0],
              [0.5, 0.0, -2.0, 0.0, 1.0, -1.0, 0.0, 2.0],
          ],
          output=[6.0, 1.0, -2.0, 5.0, 2.0, -8.0, -6.0, -3.5]
      ),
  ]
  constants = [0]
  description = ('like tf.divide(), but when dividing by 0, return the '
                 'numerator')
  target_program = 'tf.where(tf.cast(in2, tf.bool), x=tf.divide(in1, in2), y=in1)'
  source = 'https://stackoverflow.com/questions/53643339/tensorflow-overriding-tf-divide-to-return-the-numerator-when-dividing-by-0'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_24')


def stackoverflow_25():
  examples = [
      benchmark.Example(
          inputs=[
              3,
              4,
          ],
          output=[[1.0, 0.0, 0.0],
                  [0.0, 1.0, 0.0],
                  [0.0, 0.0, 1.0],
                  [1.0, 0.0, 0.0],
                  [0.0, 1.0, 0.0],
                  [0.0, 0.0, 1.0],
                  [1.0, 0.0, 0.0],
                  [0.0, 1.0, 0.0],
                  [0.0, 0.0, 1.0],
                  [1.0, 0.0, 0.0],
                  [0.0, 1.0, 0.0],
                  [0.0, 0.0, 1.0]]
      ),
  ]
  constants = []
  description = 'copy the tensor tf.eye(3), 4 times'
  target_program = 'tf.tile(tf.eye(in1), (in2, 1))'
  source = 'https://stackoverflow.com/questions/53602691/duplicate-a-tensor-n-times'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_25')


def stackoverflow_26():
  examples = [
      benchmark.Example(
          inputs=[
              [[[3, 4], [1, 2]],
               [[5, -2], [-10, 3]],
               [[10, 20], [-4, 7]]]
          ],
          output=[10, -4, 33]
      ),
  ]
  constants = []
  description = 'reduction operation for multiple dimensions simultaneously'
  target_program = 'tf.reduce_sum(tf.reduce_sum(in1, axis=1), axis=1)'
  source = 'https://stackoverflow.com/questions/54294780/how-to-perform-reduce-op-on-multiple-dimensions-at-once'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_26')


def stackoverflow_27():
  examples = [
      benchmark.Example(
          inputs=[
              [0, 3, 5, 6],
              8,
          ],
          output=[1, 0, 0, 1, 0, 1, 1, 0]
      ),
  ]
  constants = []
  description = 'boolean tensor with 1 at the indices in the input tensor'
  target_program = 'tf.cast(tf.reduce_max(tf.one_hot(in1, in2), axis=0), tf.int32)'
  source = 'https://stackoverflow.com/questions/54225704/how-do-i-get-a-tensor-representing-the-on-positions-in-the-original-tensor'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_27')


def stackoverflow_28():
  examples = [
      benchmark.Example(
          inputs=[
              [[[5, 3],
                [0, 2]],
               [[7, 4],
                [5, 1]],
               [[10, 20],
                [15, 30]],
               [[11, 16],
                [14, 12]],
               [[-2, -7],
                [-4, 6]]],
              [1, 0, 1, 1, 0],
          ],
          output=[[3, 2], [7, 5], [20, 30], [16, 12], [-2, -4]]
      ),
  ]
  constants = []
  description = 'extract columns from a 3D tensor given column indices'
  target_program = 'tf.squeeze(tf.gather(in1, tf.expand_dims(in2, 1), axis=-1, batch_dims=1))'
  source = 'https://stackoverflow.com/questions/54274074/selecting-columns-from-3d-tensor-according-to-a-1d-tensor-of-indices-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_28')


def stackoverflow_29():
  examples = [
      benchmark.Example(
          inputs=[
              [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0],
              [0.1, -10, -0.1, 1.1, 0.41],
          ],
          output=[6, 0, 5, 11, 8]
      ),
  ]
  constants = []
  description = 'place continuous values into buckets given bucket boundaries'
  target_program = "tf.searchsorted(in1, in2, side='left')"
  source = 'https://stackoverflow.com/questions/54155085/bucketing-continous-value-tensors-in-tensorflow'  # lint: NOTYPO
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_29')


def stackoverflow_30():
  examples = [
      benchmark.Example(
          inputs=[
              [[1., 2.], [3., 4.], [5., 6.]],
              [[9., 4.], [8., 5.], [7., 6.]],
          ],
          output=[[math.sqrt(68), math.sqrt(58), math.sqrt(52)],
                  [math.sqrt(36), math.sqrt(26), math.sqrt(20)],
                  [math.sqrt(20), math.sqrt(10), math.sqrt(4)]]
      ),
  ]
  constants = []
  description = 'compute Euclidean distance between two tensors'
  # StackOverflow answer is incorrect.
  target_program = 'tf.sqrt(tf.reduce_sum(tf.square(tf.subtract(tf.expand_dims(in1, 1), tf.expand_dims(in2, 0))), axis=2))'
  source = 'https://stackoverflow.com/questions/54147780/tensorflow-how-to-calculate-the-euclidean-distance-between-two-tensor'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_30')


def stackoverflow_31():
  examples = [
      benchmark.Example(
          inputs=[
              tf.SparseTensor(indices=[[0, 0], [0, 1], [1, 1]],
                              values=[1.0, 1.5, -2.0],
                              dense_shape=[2, 2]),
              [[3.0, 1.0], [0.2, -1.0]],
          ],
          output=5.29
      ),
  ]
  constants = []
  description = 'squared error between two tensors, one being a sparse tensor'
  target_program = 'tf.reduce_sum(tf.square(tf.subtract(in2, tf.sparse.to_dense(in1))))'
  source = 'https://stackoverflow.com/questions/45032668/tensorflow-how-to-compute-the-square-error-between-a-tensor-and-a-sparse-tensor'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_31')


def stackoverflow_32():
  examples = [
      benchmark.Example(
          inputs=[
              [[0.1, 0.6, 0.2, 0.1],
               [0.3, 0.1, 0.4, 0.2],
               [0.2, 0.1, 0.2, 0.5]],
          ],
          output=[1.3, 1.5, 2.0]
      ),
  ]
  constants = []
  description = 'weighted sum across rows, where the column index is the weight'
  target_program = 'tf.tensordot(in1, tf.cast(tf.range(4), tf.float32), 1)'
  source = 'https://stackoverflow.com/questions/48659449/how-to-compute-the-weighted-sum-of-a-tensor-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_32')


def stackoverflow_33():
  examples = [
      benchmark.Example(
          inputs=[
              [[.3, .1, .4], [.1, .5, .9], [.2, .6, .5], [.3, .5, .8],
               [.9, .7, .9]],
              [[.3, .2, .3], [.8, .4, .6], [.2, .6, .4], [.3, .3, .8]],
          ],
          output=[0.02, 0.19, 0.01, 0.04]
      ),
  ]
  constants = []
  description = 'find the minimum distance between two sets of points'
  target_program = 'tf.reduce_min(tf.reduce_sum(tf.square(tf.subtract(tf.expand_dims(in1, 0), tf.expand_dims(in2, 1))), axis=2), axis=1)'
  source = 'https://stackoverflow.com/questions/40558251/computing-minimum-distance-for-each-element-in-a-tensor-relative-to-another-tens'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_33')


def stackoverflow_34():
  examples = [
      benchmark.Example(
          inputs=[
              [[[1, 2], [3, 4]],
               [[5, 6], [7, 8]],
               [[10, 20], [30, 40]]],
              [3, 5, 10],
          ],
          output=[[128, 236], [344, 452]]
      ),
  ]
  constants = []
  description = 'compute a weighted sum of tensors'
  target_program = 'tf.tensordot(in2, in1, 1)'
  source = 'https://stackoverflow.com/questions/49532371/compute-a-linear-combination-of-tensors-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_34')


def stackoverflow_35():
  examples = [
      benchmark.Example(
          inputs=[
              [[[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]],
               [[10., 20.], [30., 40.], [50., 60.]]],
              [[[9.0, 8.0], [7.0, 6.0], [5.0, 4.0]],
               [[90., 80.], [70., 60.], [50., 40.]]],
              [0.1, 0.4, 0.8],
          ],
          output=[[[8.2, 7.4], [5.4, 5.2], [5.0, 5.6]],
                  [[82., 74.], [54., 52.], [50., 56.]]]
      ),
  ]
  constants = []
  description = 'linear interpolation between two tensors'
  target_program = 'tf.add(in2, tf.multiply(tf.expand_dims(in3, 1), tf.subtract(in1, in2)))'
  source = 'https://stackoverflow.com/questions/49643371/keras-compute-convex-combination-of-two-tensors'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_35')


def stackoverflow_36():
  examples = [
      benchmark.Example(
          inputs=[
              [1, 0, 1, 1, 0, 1, 0, 1],
          ],
          output=[1., 0., 0.333333, 0.25, 0., 0.166667, 0., 0.125]
      ),
  ]
  constants = []
  description = 'divide each element by the column index'
  target_program = 'tf.cast(tf.divide(in1, tf.add(in1, tf.range(8))), tf.float32)'
  source = 'https://stackoverflow.com/questions/43306788/divide-elements-of-1-d-tensor-by-the-corrispondent-index'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_36')


def stackoverflow_37():
  examples = [
      benchmark.Example(
          inputs=[
              [[[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
                [[1.2, 3.4, 5.6], [7.8, 9.8, 7.6]]]],
              [0.5, 1.0, 2.0],
          ],
          output=[[[8.5, 19.0], [15.2, 28.9]]]
      ),
  ]
  constants = []
  description = 'dot product a vector with last dimension of a tensor'
  target_program = 'tf.tensordot(in1, in2, 1)'
  source = 'https://stackoverflow.com/questions/49206051/multiply-4-d-tensor-with-1-d-tensor'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_37')


def stackoverflow_38():
  # To simplify the problem, and to get more than one number as output, this
  # doesn't include the final reduce_sum step.
  examples = [
      benchmark.Example(
          inputs=[
              [9, 2, 5, 3, 7, 4],
              [[0, 0, 1, 0, 1, 0],
               [1, 0, 0, 0, 0, 0],
               [0, 1, 1, 1, 0, 1]],
          ],
          output=[35, 9, 120]
      ),
  ]
  constants = []
  description = 'compute the product of marked elements'
  target_program = 'tf.reduce_prod(tf.maximum(tf.reduce_max(in2), tf.multiply(in1, in2)), axis=1)'
  source = 'https://stackoverflow.com/questions/49511529/tensorflow-compute-multiplication-by-binary-matrix'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_38')


def stackoverflow_39():
  examples = [
      benchmark.Example(
          inputs=[
              [[-1.5, 1.0, 0.9, 2.0],
               [1.1, 0.0, -0.1, -0.9],
               [-1.0, 0.1, -1.1, 2.5]],
          ],
          output=[[2.25, 1.0, 0.0, 4.0],
                  [1.21, 0.0, 0.0, 0.0],
                  [1.0, 0.0, 1.21, 6.25]]
      ),
  ]
  constants = []
  description = ('set to 0 the elements with absolute value less than 1, and '
                 'square the other elements')
  target_program = 'tf.multiply(tf.square(in1), tf.cast(tf.cast(tf.cast(in1, tf.int32), tf.bool), tf.float32))'
  source = 'https://stackoverflow.com/questions/37912161/how-can-i-compute-element-wise-conditionals-on-batches-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_39')


def stackoverflow_40():
  examples = [
      benchmark.Example(
          inputs=[
              [4, 5, 2, 7, 8, 6],
              [[0, 2], [0, 4], [1, 1], [1, 3], [2, 0], [2, 3]],
          ],
          output=[[0, 0, 4, 0, 5],
                  [0, 2, 0, 7, 0],
                  [8, 0, 0, 6, 0]]
      ),
  ]
  constants = []
  description = 'use the output of tf.nn.top_k to make a sparse tensor'
  target_program = 'tf.sparse.to_dense(tf.SparseTensor(tf.cast(in2, tf.int64), in1, (3, 5)))'
  source = 'https://stackoverflow.com/questions/43996831/make-a-sparse-tensor-based-on-the-output-of-tf-nn-top-k'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_40')


def stackoverflow_41():
  examples = [
      benchmark.Example(
          inputs=[
              [5, 2, 8, 2, 4, 1, 1, 0, 2, 1],
              3,
          ],
          output=[5, 2, 8, 4, 1, 1, 0, 2, 1]
      ),
  ]
  constants = []
  description = 'copy all elements except at the given index'
  target_program = 'tf.boolean_mask(in1, tf.not_equal(tf.constant(in2), tf.range(10)))'
  source = 'https://stackoverflow.com/questions/54499051/elegant-way-to-access-python-list-and-tensor-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_41')


def stackoverflow_42():
  examples = [
      benchmark.Example(
          inputs=[
              [4, 6, 2, 6, 7, 3, -3],
          ],
          output=[0, 0, 0, 0, 1, 0, 0]
      ),
  ]
  constants = []
  description = 'create a binary vector where the max element is 1'
  target_program = 'tf.cast(tf.equal(in1, tf.reduce_max(in1)), tf.int32)'
  source = 'https://stackoverflow.com/questions/54493814/binary-vector-of-max'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_42')


def stackoverflow_43():
  examples = [
      benchmark.Example(
          inputs=[
              [[12, 34, 56, 78, 90, 10],
               [99, 88, 77, 55, 44, 33],
               [-1, -2, -3, -4, -5, -6]],
              [0, 1, 1, 0, 2, 0],
          ],
          output=[12, 88, 77, 78, -5, 10]
      ),
  ]
  constants = []
  description = 'extract elements of a tensor given row indices'
  target_program = 'tf.gather_nd(tf.transpose(in1), tf.expand_dims(in2, 1), batch_dims=1)'
  source = 'https://stackoverflow.com/questions/54455169/better-way-to-access-individual-elements-in-a-tensor'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_43')


def stackoverflow_44():
  examples = [
      benchmark.Example(
          inputs=[
              [[3, 5, 2], [6, 2, 3],
               [8, 7, 1], [0, -3, 5],
               [-4, 7, 3], [2, 1, 6],
               [10, 20, 30], [4, 5, 6]],
          ],
          output=[[9, 7, 5],
                  [8, 4, 6],
                  [-2, 8, 9],
                  [14, 25, 36]]
      ),
  ]
  constants = [2]
  description = 'sum across columns for pairs of consecutive rows'
  target_program = 'tf.squeeze(tf.reduce_sum(tf.reshape(in1, (-1, 2, in1.shape[1])), axis=1))'
  source = 'https://stackoverflow.com/questions/54402389/sum-the-columns-for-each-two-consecutive-rows-of-a-tensor-of-3-dimensions'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_44')


def stackoverflow_45():
  examples = [
      benchmark.Example(
          inputs=[
              [[[12, 34], [56, 78], [23, 54], [76, 78], [42, 24]]],
              [1, 0, 1, 0, 1],
          ],
          output=[[[34, 12], [56, 78], [54, 23], [76, 78], [24, 42]]]
      ),
  ]
  constants = []
  description = 'reverse the order in the marked rows'
  target_program = 'tf.where(tf.sequence_mask(in2), x=tf.roll(in1, 1, -1), y=in1)'
  source = 'https://stackoverflow.com/questions/54337925/reverse-order-of-some-elements-in-tensorflow'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_45')


def stackoverflow_46():
  examples = [
      benchmark.Example(
          inputs=[
              [3, 4, 1],
          ],
          output=[0, 0, 0, 1, 1, 1, 1, 2]
      ),
  ]
  constants = []
  description = 'convert segment lengths to segment ids'
  target_program = 'tf.cast(tf.where(tf.sequence_mask(in1))[:, 0], tf.int32)'
  source = 'https://stackoverflow.com/questions/58652161/how-to-convert-2-3-4-to-0-0-1-1-1-2-2-2-2-to-utilize-tf-math-segment-sum'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_46')


def stackoverflow_47():
  examples = [
      benchmark.Example(
          inputs=[
              [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16],
              [[True, True, True, False, False],
               [True, True, False, False, False],
               [True, True, True, True, True],
               [True, True, True, True, False],
               [True, False, False, False, False],
               [True, True, False, False, False]],
          ],
          output=[[0, 1, 2, 0, 0],
                  [3, 4, 0, 0, 0],
                  [5, 6, 7, 8, 9],
                  [10, 11, 12, 13, 0],
                  [14, 0, 0, 0, 0],
                  [15, 16, 0, 0, 0]]
      ),
  ]
  constants = []
  description = 'put given values into a sequence mask'
  # Note: StackOverflow answer is wrong (gather index out of bounds).
  # Solution is simpler with an auxiliary variable:
  # linear_mask = tf.reshape(tf.cast(in2, tf.int32), [-1])
  # output = tf.reshape(tf.gather(in1, tf.cumsum(linear_mask, exclusive=True) * linear_mask), in2.shape)
  target_program = 'tf.reshape(tf.gather(in1, tf.cumsum(tf.reshape(tf.cast(in2, tf.int32), [-1]), exclusive=True) * tf.reshape(tf.cast(in2, tf.int32), [-1])), in2.shape)'
  source = 'https://stackoverflow.com/questions/58641546/how-can-i-put-the-sequential-values-to-the-sequence-mask'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_47')


def stackoverflow_48():
  examples = [
      benchmark.Example(
          inputs=[
              [32, 53, 45, 38, 29, 89, 64, 23],
              [38, 53, 89, 38, 32, 64],
          ],
          output=[3, 1, 5, 3, 0, 6]
      ),
  ]
  constants = []
  description = 'find the indices of all elements'
  target_program = 'tf.cast(tf.argmax(tf.cast(tf.equal(in1, tf.expand_dims(in2, 1)), tf.int32), axis=1), tf.int32)'
  source = 'https://stackoverflow.com/questions/58481332/getting-the-indices-of-several-elements-in-a-tensorflow-at-once'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_48')


def stackoverflow_49():
  examples = [
      benchmark.Example(
          inputs=[
              # Shape = [3, 1, 2, 3].
              [[[[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]],
               [[[0.8, 1.0, 0.0], [0.6, 0.4, 0.2]]],
               [[[0.9, 0.8, 0.7], [0.1, 0.2, 0.3]]]],
              [2.0, 0.5, 1.0],
          ],
          output=[[[[0.2, 0.4, 0.6], [0.8, 1.0, 1.2]]],
                  [[[0.4, 0.5, 0.0], [0.3, 0.2, 0.1]]],
                  [[[0.9, 0.8, 0.7], [0.1, 0.2, 0.3]]]]
      ),
  ]
  constants = []
  description = 'multiply tensors by scalars in a batched way'
  target_program = 'tf.transpose(tf.multiply(in2, tf.transpose(in1)))'
  source = 'https://stackoverflow.com/questions/58466562/given-a-batch-of-n-images-how-to-scalar-multiply-each-image-by-a-different-scal'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_49')


def stackoverflow_50():
  examples = [
      benchmark.Example(
          inputs=[
              5,  # Rows.
              6,  # Columns.
              3,  # Index of nonzero column.
          ],
          output=[[0, 0, 0, 1, 0, 0],
                  [0, 0, 0, 1, 0, 0],
                  [0, 0, 0, 1, 0, 0],
                  [0, 0, 0, 1, 0, 0],
                  [0, 0, 0, 1, 0, 0]]
      ),
  ]
  constants = []
  description = 'create a binary matrix where a specified column is set to one'
  target_program = 'tf.cast(tf.one_hot(tf.fill((in1,), in3), in2), tf.int32)'
  source = 'https://stackoverflow.com/questions/58537495/tensorflow-initialize-a-sparse-tensor-with-only-one-line-column-not-zero'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_50')

# A template for easy copy/pasting. Copying an existing benchmark and replacing
# parts of it will lead to a state where the benchmark is half-correct, but not
# obviously so. Copy this template instead when creating new benchmarks.
"""

def stackoverflow_NUMBER():
  examples = [
      benchmark.Example(
          inputs=[
              INPUT_1,
              INPUT_2,
          ],
          output=OUTPUT
      ),
  ]
  constants = [CONSTANTS]
  description = 'DESCRIPTION'
  target_program = 'SOLUTION_PROGRAM'
  source = 'PROBLEM_SOURCE'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='stackoverflow_NUMBER')

"""  # pylint: disable=pointless-string-statement



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Defines the Value objects used in value search."""

import abc
import collections
import functools
import itertools
import operator
import six

import tensorflow as tf
from tf_coder import tf_coder_utils
from tf_coder.value_search import value_search_utils


def cache(fn):
  def new_fn(self):
    key = fn.__name__
    if key in self.cached_info:
      return self.cached_info[key]
    result = fn(self)
    self.cached_info[key] = result
    return result
  return new_fn


@six.add_metaclass(abc.ABCMeta)
class Value(object):
  """The Value class wraps a value (a Python object) used in value search.

  This class serves the following purposes:
    * During initialization, it computes and caches information about the
      wrapped value, making it easy to filter values based on their properties.
    * It stores information about how the wrapped value was created, so that all
      Value objects can compute their corresponding Python expressions.
    * It provides a consistent way of comparing wrapped values for equality.

  Once created, a Value is treated as immutable. This means that the Value
  should not be passed to operations that modify their arguments.

  If the wrapped value is a sequence, it must not be empty, its elements must
  all have the same type, and the elements must be primitives, tf.Tensors, or
  sequences. In the nested sequence case, we also require that it is tensor-like
  (i.e., calling tf.constant() succeeds without error).

  Attributes:
    value: The value (any eligible Python object) wrapped inside this Value.
    type: Exactly `type(value)`, as an attribute for convenience.

    is_primitive: True iff `type` is one of int, list, bool, or string.
    is_dtype: True iff `value` is an instance of tf.DType.
    is_sequence: True iff `value` is an instance of list or tuple.
    is_tensor: True iff `value` an instance of tf.Tensor.
    is_sparse_tensor: True iff `value` an instance of tf.SparseTensor.

    elem_type: If `value` is a sequence, this is the type of the sequence
      elements. Otherwise, this is None.
    elem_type_is_tensor: If `value` is a sequence, this is True iff the sequence
      elements are instances of tf.Tensor. Otherwise, this is False.
    elem_type_is_sparse_tensor: If `value` is a sequence, this is True iff the
      sequence elements are instances of tf.SparseTensor. Otherwise, this is
      False.
    sequence_dtype: If `value` is a sequence of primitives or a nested sequence,
      this is the dtype of the result of calling tf.constant(). Otherwise, this
      is None.
    sequence_shape: If `value` is a sequence of primitives or a nested sequence,
      this is the shape of the result of calling tf.constant(), as a list of
      ints. Otherwise, this is None.

    dtype: If `value` is a Tensor or SparseTensor, this is its dtype. Otherwise,
      this is None.
    shape: If `value` is a Tensor or SparseTensor, this is its shape (as a list
      of ints). Othwerwise, this is None.

    cached_info: A dict containing cached computations, such as the maximum
      element in a tensor.
  """

  def __init__(self, value):
    """Initializes a Value object to contain `value`.

    Args:
      value: The actual Python object that this Value represents.

    Raises:
      ValueError: If `value` is an empty sequence, or a sequence of different
        types of elements, or anything that is too large.
    """
    self.value = value
    self.type = type(value)

    self.is_primitive = self.type in tf_coder_utils.PRIMITIVE_TYPES
    self.is_dtype = isinstance(self.value, tf.DType)
    self.is_sequence = isinstance(self.value, (list, tuple))
    self.is_tensor = isinstance(value, tf.Tensor)
    self.is_sparse_tensor = isinstance(value, tf.SparseTensor)

    self.cached_info = {}

    # Sequence-related attributes and checks.
    # Sequences include lists, tuples, and namedtuples.
    self.elem_type = None
    self.elem_type_is_tensor = False
    self.elem_type_is_sparse_tensor = False
    self.sequence_dtype = None
    self.sequence_shape = None
    if self.is_sequence:
      # Sequences must be nonempty.
      # Do not assume empty sequences will implicitly convert to False in a
      # boolean context (e.g., TensorShape does not).
      if not len(value):  # pylint: disable=g-explicit-length-test
        raise ValueError('Sequences must be nonempty.')
      self.elem_type = type(value[0])

      # Sequences must contain elements of the same type.
      # Using isinstance() here is less correct because it is not symmetric,
      # i.e., `isinstance(value[i], type(value[0]))` can be different after
      # switching `i` and `0`. Instead, we force all elements to be exactly
      # the same type.
      if not all(type(elem) is self.elem_type for elem in value):  # pylint: disable=unidiomatic-typecheck
        raise ValueError('Sequences must contain elements of the same type.')

      self.elem_type_is_tensor = isinstance(value[0], tf.Tensor)
      self.elem_type_is_sparse_tensor = isinstance(value[0], tf.SparseTensor)

      # Sequences must contain tensors, SparseTensors, primitives, or other
      # sequences.
      if (not self.elem_type_is_tensor and
          not self.elem_type_is_sparse_tensor and
          self.elem_type not in tf_coder_utils.PRIMITIVE_TYPES and
          self.elem_type not in (list, tuple)):
        raise ValueError('Sequence must contain Tensors, SparseTensors, '
                         'primitives, or other sequences.')

      # The sequence must be tensor-like (i.e., tf.constant() succeeds), unless
      # the sequence contains Tensors or SparseTensors (which may have different
      # shapes or dtypes).
      if not self.elem_type_is_tensor and not self.elem_type_is_sparse_tensor:
        try:
          as_tensor = tf.constant(self.value)
          self.sequence_shape = as_tensor.shape.as_list()
          self.sequence_dtype = as_tensor.dtype
        except (ValueError, TypeError):
          raise ValueError('Sequence is not tensor-like.')
        if self.num_elements() == 0:
          raise ValueError('Sequence represents an empty tensor.')

    # Tensor-related attributes and checks.
    self.dtype = None
    self.shape = None
    if self.is_tensor:
      self.dtype = value.dtype
      self.shape = value.shape.as_list()
      if self.num_elements() == 0:
        raise ValueError('Tensor is empty.')
      if not value_search_utils.check_tensor_size(self):
        raise ValueError('Tensor value is too large.')

    # Sparse tensor checks.
    if self.is_sparse_tensor:
      self.dtype = value.dtype
      self.shape = value.shape.as_list()

      # Check that the indices are in bounds.
      # TODO(kshi): This test is expensive (~7% slowdown). Consider only doing
      # this test when necessary, e.g., when the user provides a SparseTensor,
      # or on the result of a tf.SparseTensor(...) operation.
      bounding_box = tf.sparse.reset_shape(value)
      try:
        tf.sparse.reset_shape(bounding_box, new_shape=value.shape)
      except ValueError:
        raise ValueError('SparseTensor has out-of-bounds index.')

      if not value_search_utils.check_sparse_tensor_size(self):
        raise ValueError('SparseTensor value is too large.')
      if self.num_elements() == 0:
        raise ValueError('SparseTensor is empty.')

    # At most one of is_primitive, is_dtype, is_sequence, is_tensor, and
    # is_sparse_tensor can be True, or there is a major bug.
    assert sum((self.is_primitive, self.is_dtype, self.is_sequence,
                self.is_tensor, self.is_sparse_tensor)) <= 1

    self._repr_cache = None

  def __repr__(self):
    """Returns a string representation of the value.

    Values are considered equal if and only if their string representations (as
    computed by this function) are equal.
    """
    if self._repr_cache is None:
      self._repr_cache = tf_coder_utils.object_to_string(self.value)
    return self._repr_cache

  def __hash__(self):
    """Implements hash so that Value objects can be used as dict keys."""
    return hash(repr(self))

  def __eq__(self, other):
    """Returns whether this Value object is equal to `other`.

    Args:
      other: The other object to compare to.

    Values are considered equal if and only if their string representations (as
    computed by __repr__) are equal.
    """
    if not isinstance(other, Value):
      return NotImplemented
    return repr(self) == repr(other)

  def __ne__(self, other):
    """Returns whether this Value object is not equal to `other`."""
    return not self == other

  def copy(self):
    """Returns a copy of this Value, recursing through OperationValues."""
    raise NotImplementedError('Calling copy() on unhandled Value subclass {}'
                              .format(type(self)))

  @abc.abstractmethod
  def reconstruct_expression(self, use_cache=True):
    """Returns a code expression (as a string) that creates this value.

    This can be slow and should not be called in a tight loop.

    Args:
      use_cache: If True, the reconstruction may be looked up from a cache. If
        False, the reconstruction will be recomputed on each call.
    """

  def reconstruct_expression_with_input_names(self):
    """Returns the code expression and a set of input names that were used.

    The input names can be used to determine if a solution uses any or all of
    the inputs.

    This implementation assumes no inputs are used. Subclasses should override
    this method if they can use inputs.
    """
    return self.reconstruct_expression(), set()

  def reconstruct_all_expressions_with_input_names(self, seen_values=None):
    """Returns a list of code expressions and input names that were used.

    By default this simply calls reconstruct_expression_with_input_names().
    Subclasses should override this method if there are multiple possible
    reconstructions of the value.

    Args:
      seen_values: A set of Value objects that should not be used as
        descendants because they already appeared as an ancestor.
    """
    del seen_values  # Unused here, but used by OperationValue's implementation.
    return [self.reconstruct_expression_with_input_names()]

  @cache
  def num_elements(self):
    """Returns the number of elements in the wrapped value."""
    if self.is_sparse_tensor:
      return tf_coder_utils.num_tensor_elements(self.value.values)
    else:
      return tf_coder_utils.num_tensor_elements(self.value)

  @cache
  def max(self):
    """Returns the maximum of the wrapped value."""
    if self.elem_type in (int, float):
      return max(self.value)
    elif self.is_sparse_tensor:
      return tf_coder_utils.max_tensor_value(self.value.values)
    else:
      return tf_coder_utils.max_tensor_value(self.value)

  @cache
  def min(self):
    """Returns the minimum of the wrapped value."""
    if self.elem_type in (int, float):
      return min(self.value)
    elif self.is_sparse_tensor:
      return tf_coder_utils.min_tensor_value(self.value.values)
    else:
      return tf_coder_utils.min_tensor_value(self.value)

  @cache
  def reduce_sum(self):
    """Returns the sum of the wrapped value's elements."""
    if self.is_sequence:
      return sum(self.value)
    else:
      return float(tf.reduce_sum(tf.cast(self.value, tf.float32)))

  @cache
  def reduce_prod(self):
    """Returns the product of the wrapped value's elements."""
    if self.is_sequence:
      return functools.reduce(operator.mul, self.value, 1)
    else:
      return float(tf.reduce_prod(tf.cast(self.value, tf.float32)))

  @cache
  def numpy_tolist(self):
    """Returns the wrapped value's elements as a list."""
    return self.value.numpy().tolist()

  @cache
  def has_int_dtype(self):
    """Returns whether the wrapped value has an int dtype."""
    return self.dtype in tf_coder_utils.INT_DTYPES

  @cache
  def has_float_dtype(self):
    """Returns whether the wrapped value has a float dtype."""
    return self.dtype in tf_coder_utils.FLOAT_DTYPES

  @cache
  def is_int_dtype(self):
    """Returns whether the wrapped value is itself an int dtype."""
    return self.value in tf_coder_utils.INT_DTYPES


OperationApplication = collections.namedtuple(
    'OperationApplication', ['operation', 'arg_values'])


class OperationValue(Value):
  """A Value resulting from the application of an Operation.

  Attributes:
    operation_applications: A list of OperationApplication namedtuples that
      describe different ways of reaching the value wrapped by this Value.
  """

  def __init__(self, value, operation, arg_values):
    super(OperationValue, self).__init__(value)
    self.operation_applications = [OperationApplication(operation=operation,
                                                        arg_values=arg_values)]
    self._expression_cache = None

  def copy(self):
    """See base class."""
    copied_value = OperationValue(self.value, operation=None, arg_values=None)
    copied_value.operation_applications = []
    for operation_application in self.operation_applications:
      copied_arg_values = [arg_value.copy()
                           for arg_value in operation_application.arg_values]
      copied_value.operation_applications.append(
          OperationApplication(operation_application.operation,
                               copied_arg_values))
    return copied_value

  def reconstruct_expression(self, use_cache=True):
    """See base class."""
    if use_cache and self._expression_cache is not None:
      return self._expression_cache
    first_way = self.operation_applications[0]
    self._expression_cache = first_way.operation.reconstruct_expression(
        first_way.arg_values, use_cache=use_cache)
    return self._expression_cache

  def reconstruct_expression_with_input_names(self):
    """See base class."""
    first_way = self.operation_applications[0]
    return first_way.operation.reconstruct_expression_with_input_names(
        first_way.arg_values)

  def merge_reconstructions(self, other_operation_value):
    """Adds another way of constructing this same value.

    Args:
      other_operation_value: An OperationValue that is equal to this object, but
        with a different reconstruction.
    """
    self.operation_applications.extend(
        other_operation_value.operation_applications)

  def reconstruct_all_expressions_with_input_names(self, seen_values=None):
    """See base class."""
    if seen_values is None:
      seen_values = set()

    seen_values.add(self)

    results = []
    for operation_application in self.operation_applications:
      operation = operation_application.operation
      arg_values = operation_application.arg_values
      if any(arg_value in seen_values for arg_value in arg_values):
        continue

      # A list containing a List[Tuple[reconstruction, used input names]] for
      # each argument.
      all_arg_reconstructions = [
          arg_value.reconstruct_all_expressions_with_input_names(
              seen_values=seen_values.copy())
          for arg_value in arg_values
      ]

      for product in itertools.product(*all_arg_reconstructions):
        expressions_list, input_names_list = zip(*product)
        results.append(
            (operation.reconstruct_expression_from_strings(expressions_list),
             set.union(*input_names_list)))

    return results


class ConstantValue(Value):
  """A constant value that is not created by any operation."""

  def copy(self):
    """See base class."""
    return ConstantValue(self.value)

  def reconstruct_expression(self, use_cache=True):
    """See base class."""
    return repr(self.value)


class InputValue(Value):
  """A value provided by the user as an input.

  Attributes:
    name: The name of the "Python variable" represented by this InputValue.
  """

  def __init__(self, value, name, skip_tensor_conversion=False):
    """Initializes an InputValue to contain `value` with name `name`."""
    if (not skip_tensor_conversion and
        not isinstance(value, tf_coder_utils.PRIMITIVE_TYPES)):
      try:
        value = tf_coder_utils.convert_to_tensor(value)
      except (TypeError, ValueError):
        pass
    super(InputValue, self).__init__(value)
    self.name = name

  def copy(self):
    """See base class."""
    return InputValue(self.value, self.name, skip_tensor_conversion=True)

  def reconstruct_expression(self, use_cache=True):
    """See base class."""
    return self.name

  def reconstruct_expression_with_input_names(self):
    """See base class."""
    return self.name, {self.name}


class OutputValue(Value):
  """A Value representing the user's desired output.

  This class is simply a wrapper aound the output value so that it can be
  compared to other Value objects.
  """

  def __init__(self, value):
    """Initializes an OutputValue to contain `value`."""
    value = tf_coder_utils.convert_to_tensor(value)
    super(OutputValue, self).__init__(value)

  def reconstruct_expression(self, use_cache=True):
    """An OutputValue is not created from any expression."""
    raise NotImplementedError()


class ExpressionValue(Value):
  """A Value arising from a known simple expression."""
  # TODO(kshi): This class is currently unused (outside of being convenient for
  # tests), but it may be useful to implement expressions like `in1.shape[1]`.

  def __init__(self, value, expression, used_input_names=None):
    """Initializes the ExpressionValue."""
    super(ExpressionValue, self).__init__(value)
    self.expression = expression
    self.used_input_names = set(used_input_names) if used_input_names else set()

  def copy(self):
    """See base class."""
    return ExpressionValue(self.value, self.expression)

  def reconstruct_expression(self, use_cache=True):
    """See base class."""
    return self.expression

  def reconstruct_expression_with_input_names(self):
    """See base class."""
    return self.expression, self.used_input_names



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Settings specific to the value search approach to the TF-Coder problem."""

import ast
import os
from typing import Any, Dict, List, Text


class Settings(object):
  """Stores settings for TF-Coder's value search algorithm."""

  def __init__(self):
    # A string describing the current version of the search algorithm.
    self.algorithm_version = (
        'Value search, '
        'TF-IDF (k=5, min_score=0.15), '
        'tensor features model with F_1 loss and max weighting, '
        '2020/08/26')

    # Time limit in seconds.
    self.timeout = 60

    # Maximum number of solutions to search for.
    self.max_solutions = 1

    # Whether to only search for solutions with minimal weight.
    self.only_minimal_solutions = True

    # Maximum number of seconds to spend searching for solutions after the
    # first.
    self.max_extra_solutions_time = 10

    # Maximum weight of an expression to search for.
    self.max_weight = 300

    # Whether to require solutions to use all inputs, at least one input, or no
    # restriction.
    self.require_all_inputs_used = True
    self.require_one_input_used = True

    # The description handler to use.
    self.description_handler_name = 'tfidf_5_0.15'

    # Other settings organized into separate objects.
    self.operations = OperationSettings()
    self.tensor_model = TensorModelSettings()
    self.printing = PrintSettings()
    self.paper_experiments = PaperExperimentSettings()

  # Used to parse setting names.
  _GROUP_NAMES = ['operations', 'tensor_model', 'printing', 'paper_experiments']

  def set(self, name: Text, value: Any) -> None:
    """Sets the setting with the given name to the given value.

    Args:
      name: The name of the setting to set. For example, 'timeout' is used to
        set `self.timeout`, and either 'printing.statistics' or
        'printing_statistics' can be used to set `self.printing.statistics`.
      value: The value to set the setting to.
    """
    if hasattr(self, name):
      setattr(self, name, value)
    else:
      for group_name in Settings._GROUP_NAMES:
        if (name.startswith(group_name) and
            name[len(group_name)] in {'.', '_'}):
          reduced_name = name[len(group_name) + 1 : ]
          group = getattr(self, group_name)
          if hasattr(group, reduced_name):
            setattr(group, reduced_name, value)
            break
      else:
        raise ValueError(
            'The name `{}` does not match any setting.'.format(name))

  def as_dict(self) -> Dict[Text, Any]:
    """Returns all settings as a dict."""
    result = {}
    for name, value in self.__dict__.items():
      if name in Settings._GROUP_NAMES:
        for inner_name, inner_value in value.__dict__.items():
          full_name = name + '.' + inner_name
          result[full_name] = inner_value
      else:
        result[name] = value
    return result


class OperationSettings(object):
  """Settings about operations to use during search."""

  def __init__(self):
    # Whether to limit sparse operations to benchmarks that contain
    # SparseTensors in their examples.
    self.limit_sparse_operations = True

    # TODO(kshi): Add options to exclude specific operations, or prioritize
    # user-chosen operations.


class TensorModelSettings(object):
  """Settings for the tensor features model."""

  def __init__(self):
    # This model is trained with max-weight F_1 loss, learning rate 1e-4, and 1
    # hidden layer of size 1024.
    model_dir = os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)),
                     '..', 'models', 'trained_model'))
    self.checkpoint_path = os.path.join(model_dir, 'ckpt-1172')
    self.config_path = os.path.join(model_dir, 'config.json')

    # How the model should prioritize or deprioritize operations.
    self.prioritize_threshold = 0.5
    self.prioritize_multiplier = 0.75
    self.deprioritize_threshold = -1  # No deprioritization.
    self.deprioritize_multiplier = 1.25
    self.max_deprioritized = 100


class PrintSettings(object):
  """Settings that affect printing to stdout."""

  def __init__(self):
    # Whether to print intermediate results and progress. Setting this to True
    # will cause significant slowdown from computing and printing many
    # expressions.
    self.verbose = False

    # Whether to print every FunctionOperation application before it occurs.
    # Setting this to True will cause a huge amount of output and significant
    # slowdown.
    self.all_apply = False

    # Whether to print warnings about too-large tensors.
    self.tensor_size_warnings = False

    # Whether to print progress at each iteration of target expression weight.
    self.progress = False

    # Whether to print bad solutions.
    self.bad_solutions = False

    # Whether to print statistics about operations and executions.
    self.statistics = False

    # Whether to print statistics sorted by time (versus by name). Ignored if
    # `statistics` is False.
    self.statistics_sort_by_time = True

    # Whether to print the operations that are prioritized or deprioritized.
    self.prioritized_operations = False
    self.deprioritized_operations = False


class PaperExperimentSettings(object):
  """Settings for experiments in the PLDI 2020 paper."""

  def __init__(self):
    self.skip_filtering = False
    self.uniform_weights = False


def default_settings() -> Settings:
  """Returns a Settings object with default settings."""
  return Settings()


def from_dict(overrides: Dict[Text, Any]) -> Settings:
  """Sets settings using a dict to override defaults."""
  settings = default_settings()
  for name, value in overrides.items():
    settings.set(name, value)
  return settings


def from_list(overrides: List[Text]) -> Settings:
  """Sets settings using a list to override defaults.

  Args:
    overrides: A list of strings like 'timeout=120' or
      'printing.statistics=True'. Each string should contain exactly one '='
      character. The portion before the '=' character names a setting to
      override. The portion after the '=' character describes the value of the
      setting, in a form parseable by ast.literal_eval().

  Raises:
    ValueError: If any element of `overrides` cannot be processed
      successfully.

  Returns:
    A Settings object.
  """
  settings = default_settings()
  for override_string in overrides:
    if override_string.count('=') != 1:
      raise ValueError("The override string {!r} does not contain exactly "
                       "one '=' character.".format(override_string))
    equals_index = override_string.index('=')
    name = override_string[:equals_index]
    value_string = override_string[equals_index + 1:]
    try:
      value = ast.literal_eval(value_string)
      settings.set(name, value)
    except Exception as e:
      raise ValueError(
          'Exception raised in ast.literal_eval on {!r}: {}'.format(
              value_string, e))
  return settings



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Collects all known TF-Coder benchmarks."""

import inspect

from tf_coder.benchmarks import autopandas_benchmarks
from tf_coder.benchmarks import google_benchmarks
from tf_coder.benchmarks import simple_benchmarks
from tf_coder.benchmarks import stackoverflow_benchmarks
from tf_coder.benchmarks import test_benchmarks


_ALL_BENCHMARK_MODULES = [test_benchmarks, simple_benchmarks, google_benchmarks,
                          stackoverflow_benchmarks, autopandas_benchmarks]


def all_benchmarks(include_ignored=False, modules=None):
  """Returns a list of all benchmarks.

  Args:
    include_ignored: A boolean indicating whether the search should include
      ignored benchmarks.
    modules: A list of module objects to inspect for benchmark functions. If
      None, all known relevant modules are used.

  Returns:
    A list of benchmark.Benchmark objects.
  """
  if modules is None:
    modules = _ALL_BENCHMARK_MODULES
  members = sum((inspect.getmembers(benchmark_module, inspect.isfunction)
                 for benchmark_module in modules), [])
  benchmark_list = []
  for unused_name, benchmark_function in members:
    benchmark = benchmark_function()
    if include_ignored or not benchmark.should_ignore:
      benchmark_list.append(benchmark)
  return benchmark_list


def find_benchmark_with_name(benchmark_name, include_ignored=False,
                             modules=None):
  """Returns a benchmark with the given name.

  Args:
    benchmark_name: A name (string) to search for.
    include_ignored: A boolean, used as described in all_benchmarks().
    modules: A list of module objects, used as described in all_benchmarks(). If
      None, all known relevant modules are used.

  Returns:
    A benchmark.Benchmark with the given name, if there is exactly one such
    benchmark. If there are zero or multiple such benchmarks, None is returned.
  """
  benchmark_list = all_benchmarks(include_ignored=include_ignored,
                                  modules=modules)
  matching_benchmarks = [benchmark
                         for benchmark in benchmark_list
                         if benchmark.name == benchmark_name]
  if len(matching_benchmarks) == 1:
    return matching_benchmarks[0]
  return None


def get_chosen_benchmarks(benchmark_name, include_ignored=False, modules=None):
  """Returns benchmarks according to the benchmark_name argument.

  Args:
    benchmark_name: The string name of a desired benchmark, or "ALL".
    include_ignored: A boolean, used as described in all_benchmarks().
    modules: A list of module objects, used as described in all_benchmarks(). If
      None, all known relevant modules are used.

  Returns:
    A list of benchmark.Benchmark objects.
  """
  if benchmark_name == 'ALL':
    return all_benchmarks(modules=modules)
  benchmark = find_benchmark_with_name(
      benchmark_name, include_ignored=include_ignored, modules=modules)
  if benchmark is None:
    return []
  return [benchmark]



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for operation_filtering.py."""

from absl import logging
from absl.testing import absltest
import tensorflow as tf
from tf_coder import filter_group
from tf_coder import tf_functions
from tf_coder.value_search import operation_base
from tf_coder.value_search import operation_filtering
from tf_coder.value_search import value


class DummyOperation(operation_base.Operation):
  """A dummy subclass of Operation used for testing."""

  def __init__(self, num_args, function_info):
    """Creates a DummyOperation with the given arity and FunctionInfo."""
    metadata = operation_base.OperationMetadata(docstring='dummy operation')
    super(DummyOperation, self).__init__(num_args, weight=1, metadata=metadata)
    self.function_info = function_info

  def _compute_name(self):
    return self.function_info.name

  def apply(self, arg_values):
    return None

  def reconstruct_expression_from_strings(self, arg_strings):
    return 'dummy operation'


class OperationFilteringTest(absltest.TestCase):

  def test_get_type_filter(self):
    int_type_filter = operation_filtering.get_type_filter(int)
    self.assertTrue(int_type_filter(value.ConstantValue(1)))
    self.assertFalse(int_type_filter(value.ConstantValue(1.0)))

  def test_get_dtype_filter(self):
    int32_dtype_filter = operation_filtering.get_dtype_filter(tf.int32)
    self.assertTrue(int32_dtype_filter(value.ConstantValue(tf.constant(1))))
    self.assertFalse(int32_dtype_filter(value.ConstantValue(tf.constant(1.0))))

  def test_get_dtype_filter_raises_for_non_dtype(self):
    with self.assertRaises(TypeError):
      operation_filtering.get_dtype_filter(int)

  def test_add_filters_to_function_operation_handles_all_filter_groups(self):
    for group in filter_group.FilterGroup:
      if group == filter_group.FilterGroup.NONE:
        arity = 1
      else:
        arity = int(group.name[group.name.rfind('_') + 1:])

      dummy_operation = DummyOperation(arity,
                                       tf_functions.FunctionInfo(
                                           name='tf.dummy()',
                                           filter_group=group,
                                           weight=1))
      operation_filtering.add_filters_to_function_operation(dummy_operation)

      if group is filter_group.FilterGroup.NONE:
        self.assertIsNone(dummy_operation._value_filters_list)
        self.assertIsNone(dummy_operation._apply_filter)
      else:
        self.assertTrue(dummy_operation._value_filters_list is not None or
                        dummy_operation._apply_filter is not None)

  def test_add_filters_to_function_operation_raises_for_unknown_group(self):
    dummy_operation = DummyOperation(1,
                                     tf_functions.FunctionInfo(
                                         name='tf.dummy()',
                                         filter_group='UNKNOWN',
                                         weight=1))
    with self.assertRaises(ValueError):
      operation_filtering.add_filters_to_function_operation(dummy_operation)


if __name__ == '__main__':
  logging.set_verbosity(logging.ERROR)

  absltest.main()
